#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re


def string_unquote(string):
    quote = string[:1]
    if quote != string[-1:] or ord(quote) not in (0x22, 0x27):
        raise ValueError(F'not a valid quoted string: {string}')
    string = string[1:-1].replace(quote + quote, quote)
    if quote == '"':
        string = string.replace('`"', '"')
    return string


def string_quote(string):
    return '"{}"'.format(string.replace('"', '""'))


def string_escape(string):
    def escaper(match):
        char = match.group(1)
        return '`' + {
            '\0': '0',
            '\a': 'a',
            '\b': 'b',
            '\f': 'f',
            '\n': 'n',
            '\r': 'r',
            '\t': 't',
            '\v': 'v',
        }.get(char, char)
    escaped = re.sub(R'(?<!`)([\x00\x07-\x0D`])', escaper, string)
    return re.sub(R'(?<!`)\$(?![\w\(\{\$\?\^:])', '`$', escaped)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re

from ....lib.patterns import formats
from .. import Deobfuscator
from . import string_unquote, string_quote, string_escape


class deob_ps1_stringreplace(Deobfuscator):
    def deobfuscate(self, data):
        def dash_replace(match):
            string, case, needle, insert = match.groups()
            case = '(?i)' if 'c' not in case else ''
            string = string_unquote(string)
            needle = string_unquote(needle)
            # transforming "Tvar".replace('T', '$')
            # into "$var" would trigger a variable substitution,
            # hence we need to escape the insert variable:
            insert = string_escape(string_unquote(insert))
            needle = re.escape(needle)
            return string_quote(re.sub(needle, lambda _: insert, string))

        def dot_replace(match):
            string, needle, insert = match.groups()
            string = string_unquote(string)
            needle = string_unquote(needle)
            insert = string_escape(string_unquote(insert))
            return string_quote(string.replace(needle, insert))

        data = re.sub(
            R'({s})\s*-([ci]?)replace\s*({s})\s*,\s*({s})'.format(s=formats.ps1str),
            dash_replace,
            data,
            flags=re.IGNORECASE
        )

        data = re.sub(
            R'({s}).Replace\(\s*({s})\s*,\s*({s})\s*\)'.format(s=formats.ps1str),
            dot_replace,
            data,
            flags=re.IGNORECASE
        )

        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import sys
import textwrap

from math import log

from . import HexViewerMixin, get_terminal_size
from .. import Unit

try:
    import numpy

    def _np_entropy(data: bytearray) -> float:
        value, counts = numpy.unique(data, return_counts=True)
        probs = counts / len(data)
        # 8 bits are the maximum number of bits of information in a byte
        return 0.0 + -sum(p * log(p, 2) for p in probs) / 8.0

except ImportError:
    _np_entropy = None


def entropy(data: bytearray) -> float:
    """
    Computes the entropy of `data` over the alphabet of all bytes.
    """
    if not data:
        return 0.0
    if _np_entropy:
        return _np_entropy(data)
    else:
        from collections import defaultdict
        histogram = defaultdict(int)
        for b in data:
            histogram[b] += 1
        p = 1. / len(data)
        S = [histogram[b] * p for b in histogram]
        return 0.0 + -sum(q * log(q, 2) for q in S) / 8.0


class peek(Unit, HexViewerMixin):
    """
    The unit extracts preview information of the input data and displays it on
    the standard error stream. If the standard output of this unit is connected
    by a pipe, the incoming data is forwarded. However, if the unit outputs to
    a terminal, the data is discarded instead.
    """

    def interface(self, argp):
        from ...lib.argformats import number
        lines = argp.add_mutually_exclusive_group()
        lines.add_argument('-l', '--lines', metavar='N', type=number, default=10,
            help='Specify number N of lines in the preview, default is 10.')
        lines.add_argument('-a', '--lines-all', action='store_const', dest='lines', const=None,
            help='Output all possible preview lines without restriction')
        peek = argp.add_mutually_exclusive_group()
        peek.add_argument('-d', '--decode', action='store_true',
            help='Attempt to decode and display printable data.')
        peek.add_argument('-e', '--esc', action='store_true',
            help='Always peek data as string, escape characters if necessary.')
        peek.add_argument('-b', '--brief', action='store_true',
            help='One line peek, implies --lines=0.')
        return super().interface(self.hexviewer_interface(argp))

    def process(self, data):
        for line in self._peeklines(data):
            print(line, file=sys.stderr)
        if not sys.stdout.isatty():
            self.log_info('forwarding input to next unit')
            yield data

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.separate = True

    def _peeklines(self, data):
        import codecs
        import string
        import itertools
        from ...lib.tools import format_size
        from ...lib.magic import magicparse

        peeks = [
            F'{entropy(bytearray(data)) * 100:05.2f}% entropy',
            format_size(len(data), align=not self.args.lines)
        ]

        magic = magicparse(data)

        if magic is not None:
            peeks.append(magic)

        header = ', '.join(peeks)

        dump = None
        working_codec = None

        if data and not self.args.brief:
            if self.args.decode:
                for codec in ('UTF8', 'UTF-16LE', 'UTF-16', 'UTF-16BE'):
                    try:
                        self.log_info(F'trying to decode as {codec}.')
                        handler = 'backslashreplace' if self.args.esc else 'strict'
                        decoded = codecs.decode(data, codec, errors=handler)
                        count = sum(x in string.printable for x in decoded)
                        ratio = count / len(data)
                    except ValueError as V:
                        self.log_info('decoding failed:', V)
                        continue
                    if ratio < 0.8 or any(x in decoded for x in '\b\v'):
                        self.log_info(F'data contains {ratio * 100:.2f}% printable characters, this is too low.')
                        continue

                    width = self.args.width or get_terminal_size()
                    if self.args.lines is not None:
                        decoded = decoded[:width * self.args.lines]
                    dump = [
                        line.rstrip('\r')
                        for chunk in textwrap.wrap(
                            decoded,
                            width,
                            break_on_hyphens=False,
                            replace_whitespace=False
                        )
                        for line in chunk.split('\n')
                    ]
                    working_codec = codec
                    break

            if not dump:
                total = self.args.lines * get_terminal_size() // 3
                dump = self.hexdump(data, total=total)

            dump = list(itertools.islice(dump, 0, self.args.lines))

        termsize = get_terminal_size()

        def separator(title=None):
            if title is None or termsize <= len(title) + 8:
                return termsize * '-'
            return F'--{title}' + '-' * (termsize - len(title) - 2)

        yield separator()

        if termsize:
            yield from textwrap.wrap(header, termsize)
        else:
            yield header

        if dump:
            yield separator(F'CODEC={working_codec}' if working_codec else 'HEXDUMP')
            yield from dump

        if self.separate:
            yield separator()

    def filter(self, inputs):
        from ...lib.tools import lookahead
        for last, item in lookahead(inputs):
            self.separate = last
            yield item



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from ...lib.argformats import utf8
from . import RegexUnit, TransformSubstitutionFactory


class rex(RegexUnit):
    """
    A binary grep which can apply a transformation to each match. Each match is an
    individual output and standard forking settings apply. Two additional special
    multibin handlers are available for regular expressions:
    """

    def interface(self, argp):
        argp = super().interface(argp)
        argp.add_argument(
            dest='transformations',
            metavar='transformation',
            type=utf8,
            nargs='*',
            default=None,
            help=(
                'An optional sequence of transformations to be applied to each match. '
                'Each transformation produces one output in the order in which they '
                'are given. The default transformation is $0, i.e. the entire match. '
            )
        )
        return argp

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.transforms = [TransformSubstitutionFactory(f) for f in self.args.transformations]
        if self.log_debug():
            self.log_debug('regular expression:', self.args.regex)
            for k, transform in enumerate(self.args.transformations, 1):
                self.log_debug(F'transformation {k}:', transform)

    def process(self, data):
        yield from self.matches_processed(data, self.args.regex, transforms=self.transforms)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import codecs

from .. import Unit


class recode(Unit):
    """
    Expects input string data encoded in the `from` encoding and encodes it in
    the `to` encoding, then outputs the result.
    """

    def interface(self, argp):
        argp.add_argument(metavar='from', dest='decode', type=str,
            help='input encoding of the data')
        argp.add_argument(metavar='to', dest='encode', nargs='?', default=self.codec, type=str,
            help=F'output encoding, default is {self.codec}')
        return super().interface(argp)

    def process(self, data):
        return codecs.encode(
            codecs.decode(
                data,
                self.args.decode,
                errors='surrogateescape'
            ),
            self.args.encode,
            errors='surrogateescape'
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from ... import Unit
from ....lib.argformats import number


class rot(Unit):
    """
    Rotate the characters of the alphabet by the given amount. The default
    amount is 13, providing the common (and weak) string obfuscation method.
    """

    def interface(self, argp):
        argp.add_argument('amount', nargs='?', default=13, type=number[1:25], help='rotation amount')
        return super().interface(argp)

    def process(self, data):
        def rotate(char):
            if 0x41 <= char <= 0x5A:  # A-Z
                return (char - 0x41 + self.args.amount) % 26 + 0x41
            if 0x61 <= char <= 0x7A:  # a-z
                return (char - 0x61 + self.args.amount) % 26 + 0x61
            return char
        return bytes(rotate(char) for char in data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re

from .. import Deobfuscator


class deob_vba_comments(Deobfuscator):
    def deobfuscate(self, data):
        return re.sub(
            R"^\s{0,20}(?:'|rem\b|dim\b).*$\n\r?",
            '',
            data,
            flags=re.MULTILINE | re.IGNORECASE
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import pefile
from fnmatch import fnmatch

from ... import Unit


class rsrc:
    RESOURCE_TYPES = {
        'CURSOR': 1,
        'BITMAP': 2,
        'ICON': 3,
        'MENU': 4,
        'DIALOG': 5,
        'STRING': 6,
        'FONTDIR': 7,
        'FONT': 8,
        'ACCELERATOR': 9,
        'RCDATA': 10,
        'MESSAGETABLE': 11,
        'VERSION': 16,
        'DLGINCLUDE': 17,
        'PLUGPLAY': 19,
        'VXD': 20,
        'ANICURSOR': 21,
        'ANIICON': 22,
        'HTML': 23,
        'MANIFEST': 24,
    }

    @staticmethod
    def to_name(number):
        for key in rsrc.RESOURCE_TYPES:
            if rsrc.RESOURCE_TYPES[key] == number:
                return key
        return str(number)

    def __init__(self, path):
        self.root = False
        self.needles = tuple(
            self._parse(k, x)
            for k, x in enumerate(path.split('/'))
        )

    def relax(self):
        return rsrc(F'*/{self}')

    def __str__(self):
        return '/'.join(self.to_name(x) for x in self.needles)

    def _parse(self, index, s):
        if not index:
            try:
                translation = self.RESOURCE_TYPES[s]
            except KeyError:
                pass
            else:
                self.root = True
                return translation
        try:
            return int(s, 0)
        except ValueError:
            pass
        return s


class perc(Unit):
    """
    Extract PE file resources.
    """

    def interface(self, argp):
        argp.add_argument(metavar='entry', dest='entries', type=rsrc, nargs='*',
            help='A resource path for the resource to be extracted. May contain numeric '
                 'literals and name strings with wildcards.')
        return super().interface(argp)

    def _match(self, level, needle, e):
        if e.id == needle or str(e.id) == needle:
            return True
        if level == 0 and rsrc.to_name(needle) == str(e.name):
            return True
        elif fnmatch(str(e.name), str(needle)):
            return True
        return False

    def _search(self, pe, directory, needles, level=0, path=('.rsrc',)):
        try:
            needle = needles[level]
        except IndexError:
            if level >= 3:
                self.log_warn(F'unexpected resource tree level {level + 1:d}')
            needle = '*'

        for e in directory.entries:
            if self._match(level, needle, e):
                id_string = e.id if level else rsrc.to_name(e.id)
                new_path = path + (str(e.name if e.name else id_string),)
                if e.struct.DataIsDirectory:
                    yield from self._search(pe, e.directory, needles, level + 1, new_path)
                elif needles:
                    yield pe.get_data(e.data.struct.OffsetToData, e.data.struct.Size)
                else:
                    yield '/'.join(new_path[1:]).encode('UTF8')

    def process(self, data):
        pe = pefile.PE(data=data)
        if not self.args.entries:
            yield from self._search(pe, pe.DIRECTORY_ENTRY_RESOURCE, [])
            return
        for arg in self.args.entries:
            while len(arg.needles) <= 3:
                self.log_info('searching pattern:', str(arg))
                yield from self._search(pe, pe.DIRECTORY_ENTRY_RESOURCE, arg.needles)
                arg = arg.relax()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from .. import Unit, RefineryCriticalException


class sorted(Unit):
    """
    Sorts all elements of the input `refinery.lib.frame` lexicographically.
    This unit is a `refinery.nop` on single inputs.
    """

    def filter(self, chunks):
        sortbuffer = []
        invisibles = {}

        for k, chunk in enumerate(chunks):
            if not chunk.visible:
                r = k - len(invisibles)
                invisibles.setdefault(r, [])
                invisibles[r].append(chunk)
            else:
                sortbuffer.append(chunk)

        sortbuffer.sort()

        if not invisibles:
            yield from sortbuffer
            return

        for r, chunk in enumerate(sortbuffer):
            if r in invisibles:
                yield from invisibles[r]
                del invisibles[r]
            yield chunk

        if invisibles:
            yield from invisibles[r]
            del invisibles[r]

        if invisibles:
            raise RefineryCriticalException(
                'for unknown reasons, invisible chunks were lost during '
                'the sorting process.'
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from . import PatternExtractorBase


def _lps(maxlen):
    """
    Brute force regular expression pattern for a length prefixed domain name component.
    """
    return BR'|'.join(BR'\x%02x[a-z0-9\-\_]{%d}' % (d, d) for d in range(1, maxlen + 1))


class dnsdomain(PatternExtractorBase):
    """
    Extracts domain names in the format as they appear in DNS requests. This
    can be used as a quick and dirty way to extract domains from PCAP files,
    for example.
    """

    _DOMAIN_CHARACTERS = (
        B'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
        B'abcdefghijklmnopqrstuvwxyz'
        B'0123456789-_'
    )

    _DOMAIN_PATTERN = BR'(?:%s){1,20}(?:%s)\b' % (_lps(0xFF), _lps(25))

    def interface(self, argp):
        super().interface(argp)
        return argp

    def process(self, data):
        def prefilter(matches):
            for offset, match in matches:
                match = bytearray(match)
                pos = 0
                while pos < len(match):
                    length = match[pos]
                    match[pos] = 0x2E
                    if len(match) < length + pos:
                        break
                    if any(x not in self._DOMAIN_CHARACTERS for x in match[pos + 1 : pos + length]):
                        break
                    pos += 1 + length
                else:
                    yield offset, bytes(match[1:])
        yield from self.matches_finalize(
            prefilter(self.matches_filtered(data, self._DOMAIN_PATTERN)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from .. import Unit


class cupper(Unit):
    """
    Transforms the input data to uppercase.
    """
    def process(self, data):
        return data.upper()



# -*- coding: utf-8 -*-
"""
These definitions in this module are used across the other modules in the .NET
parsing library.
"""
import struct
import datetime
import time
import base64

from io import BytesIO
from future.utils import with_metaclass


class RepresentedByNameOnly(type):
    def __repr__(self):
        return self.__name__

    def __init__(cls, name, bases, nmspc):
        def representation(self):
            return repr(self.__class__)
        setattr(cls, '__repr__', representation)


class TimeZone_UTC(datetime.tzinfo):
    def utcoffset(self, dt):
        return datetime.timedelta(0)

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return datetime.timedelta(0)


class TimeZone_Local(datetime.tzinfo):
    def __init__(self):
        self._stdoffset = datetime.timedelta(seconds=-time.timezone)
        self._altoffset = datetime.timedelta(seconds=-time.altzone)
        self._dstoffset = self._altoffset if time.daylight else self._stdoffset

    def _is_dst(self, dt):
        tt = (dt.year, dt.month, dt.day, dt.hour,
              dt.minute, dt.second, dt.weekday(), 0, 0)
        tt = time.localtime(time.mktime(tt))
        return tt.tm_isdst > 0

    def utcoffset(self, dt):
        return self._dstoffset if self._is_dst(dt) else self._stdoffset

    def dst(self, dt):
        return self.utcoffset(dt) - self._stdoffset

    def tzname(self, dt):
        return time.tzname[self._is_dst(dt)]


class MetaBox(with_metaclass(RepresentedByNameOnly)):
    pass


class Box(dict, MetaBox):
    def __init__(self, **kw):
        super(Box, self).__init__()
        for key, value in kw.items():
            setattr(self, key, value)

    def __getattr__(self, key):
        try:
            return super(Box, self).__getattribute__(key)
        except AttributeError:
            if key in self:
                return self[key]
            else:
                raise

    def __setattr__(self, name, value):
        if name.startswith('_') or hasattr(self.__class__, name):
            return super(Box, self).__setattr__(name, value)
        return self.__setitem__(name, value)


class ParserException(Exception):
    pass


class ParserEOF(ParserException):
    def __init__(self, size, data):
        ParserException.__init__(
            self,
            'attempted to read {} bytes from reader and '
            'got only {}.'.format(size, len(data))
        )
        self.data = data
        self.size = size


def unpack(item):
    while True:
        try:
            value = item.Value
        except AttributeError:
            break
        if type(value) == property:
            break
        item = value
    return item


class StreamReader(BytesIO):

    def read(self, size=None):
        if size is None:
            start = self.tell()
            self.seek(0, 2)
            size = self.tell() - start
            self.seek(start)
        data = BytesIO.read(self, size)
        if len(data) != size:
            raise ParserEOF(size, data)
        return data

    def skip(self, count):
        self.seek(count, 1)

    def align(self, blocksize):
        skip = self.tell() % blocksize
        if skip:
            self.skip(blocksize - skip)

    def __len__(self):
        pos = self.tell()
        self.seek(0, 2)
        size = self.tell()
        self.seek(pos)
        return size

    def checkpoint(self):

        class streamframe:
            def __enter__(sf):
                self.rewind = self.tell()
                return sf

            def __exit__(sf, type, value, tb):
                self.seek(self.rewind)
                return False

        return streamframe()

    def expect(self, parser, **kw):
        return unpack(self.expect_with_meta(parser, **kw))

    def expect_with_meta(self, parser, **kw):
        return parser(self, **kw)


class Blob(with_metaclass(RepresentedByNameOnly)):
    def __init__(self, reader, size=None):
        self._size = 0
        if size is None:
            start = reader.tell()
            size = self._readLengthPrefix(reader)
            self._size = reader.tell() - start
        self._data = self._consume(reader, size)
        self._size += len(self._data)

    def _readLengthPrefix(self, reader):
        size = reader.expect(Byte)
        if not size & 0x80:
            return size
        elif not size & 0x40:
            size = size & 0x3f
            size = size << 8 | reader.expect(Byte)
            return size
        elif not size & 0x20:
            size = size & 0x1f
            size = size << 8 | reader.expect(Byte)
            size = size << 8 | reader.expect(Byte)
            size = size << 8 | reader.expect(Byte)
            return size
        else:
            self._raise('length prefix invalid')

    def _consume(self, reader, size):
        return reader.read(size)

    def _raise(self, msg):
        raise ParserException(
            'attempted to parse {}: {}.'.format(repr(self), msg))

    def __len__(self):
        return self._size

    def __bytes__(self):
        return self._data


class RawBytes(Blob):
    @property
    def Value(self):
        return self._data


class StringPrimitive(Blob):
    def __init__(self, reader, size=None, align=1, codec='latin-1'):
        Blob.__init__(self, reader, size)
        self._codec = codec
        if align > 1:
            excess = self._size % align
            if excess:
                self._size += len(reader.read(align - excess))

    @property
    def Value(self):
        try:
            return self._data.decode(self._codec).rstrip(u'\x00')
        except UnicodeDecodeError:
            codec = 'utf-16le' if self._data.count(B'\0') == len(self._data) // 2 else 'latin-1'
        try:
            return self._data.decode(codec)
        except UnicodeDecodeError:
            return self._data.decode('UNICODE_ESCAPE')


class UnicodeString(StringPrimitive):
    def __init__(self, reader):
        StringPrimitive.__init__(self, reader, align=1, codec='utf-16LE')
        if len(self._data) > 0:
            if not len(self._data) % 2:
                raise ParserException('unicode string has no terminator')
            self._data = self._data[:-1]


class EncodedInteger(Blob):
    def __init__(self, reader):
        Blob.__init__(self, reader, 0)
        self._size = 0
        data = bytearray()
        value = 0
        for position in range(5):
            byte = reader.expect(Byte)
            data.append(byte)
            self._size += 1
            value |= (byte & 0b01111111) << (position * 7)
            if byte & 0b10000000 == 0:
                break
        else:
            self._raise('length prefix overflow')
        self._data = bytes(data)
        self.Value = value


class LengthPrefixedString(StringPrimitive):
    def __init__(self, reader, codec='UTF8'):
        StringPrimitive.__init__(self, reader, size=None, align=1, codec=codec)

    def _readLengthPrefix(self, reader):
        return reader.expect(EncodedInteger)


class StringGUID(Blob):
    def __init__(self, reader):
        Blob.__init__(self, reader, 16)

    @property
    def Value(self):
        rest = base64.b16encode(self._data[10:]).decode('ascii')
        values = struct.unpack('<IHHBB', self._data[:10]) + (rest,)
        return '{:08X}-{:04X}-{:04X}-{:02X}{:02X}-{}'.format(*values)

    def __str__(self):
        return self.Value


class NullTerminatedString(StringPrimitive):
    def __init__(self, reader, align=1, codec='latin-1'):
        StringPrimitive.__init__(self, reader, 0, align, codec=codec)

    def _consume(self, reader, size):
        data = b''
        assert size == 0
        while not data.endswith(b'\0'):
            data += reader.read(1)
        return data


class FixedSize(Blob):
    format = None

    def __init__(self, reader, fmt=None):
        fmt = fmt or self.format
        assert fmt is not None, 'not format specified for this FixedSize instance'
        Blob.__init__(self, reader, struct.calcsize(fmt))
        self.__value = struct.unpack('<' + fmt, self._data)
        if len(self.__value) == 0:
            self.__value = None
            return
        if len(self.__value) == 1:
            self.__value = self.__value[0]
        if hasattr(self, 'parser'):
            try:
                self.__value = self.parser(self.__value)
            except Exception as e:
                self._raise(str(e))

    @property
    def Value(self):
        return self.__value


class TypeCode(FixedSize):
    format = 'B'
    lookup = {}

    def parser(self, x):
        assert x in self.lookup, 'unknown {}({}) encountered'.format(repr(self), x)
        return self.lookup[x]


class BinaryArrayTypeEnumeration(TypeCode):
    lookup = {
        0: 'Single',
        1: 'Jagged',
        2: 'Rectangular',
        3: 'SingleOffset',
        4: 'JaggedOffset',
        5: 'RectangularOffset'
    }


class Boolean(FixedSize):
    format = 'B'
    parser = bool


class Byte(FixedSize):
    format = 'B'


class Char(FixedSize):
    format = 'b'
    parser = chr


class Int16(FixedSize):
    format = 'h'


class Int32(FixedSize):
    format = 'i'


class Int64(FixedSize):
    format = 'q'


class SByte(FixedSize):
    format = 'B'


class Single(FixedSize):
    format = 'f'


class Double(FixedSize):
    format = 'd'


class TimeSpan(FixedSize):
    format = 'Q'
    def parser(x): return datetime.timedelta(microseconds=0.1 * x)


class DateTime(FixedSize):
    format = 'Q'
    @classmethod
    def parser(x):
        hi_byte = x >> 56
        lo_part = x & 0xFFFFFFFFFFFFFFFF
        kind = hi_byte & 0b11
        time = (hi_byte >> 2) << 56 | lo_part
        assert kind < 3, 'invalid date kind'
        return datetime.datetime.fromtimestamp(time, {
            0: None,
            1: TimeZone_UTC(),
            2: TimeZone_Local()}[kind])


class UInt16(FixedSize):
    format = 'H'


class UInt32(FixedSize):
    format = 'I'


class UInt64(FixedSize):
    format = 'Q'


class Null(FixedSize):
    format = ''


class Struct(Box):

    def expect(self, parser, **kw):
        return unpack(self.expect_with_meta(parser, **kw))

    def expect_with_meta(self, parser, **kw):
        package = self._reader.expect_with_meta(parser, **kw)
        self._data += package._data
        return package

    def parse(self):
        pass

    def __init__(self, reader, cleanup=True, **kw):
        Box.__init__(self, **kw)
        self._reader = reader
        self._data = b''
        self.parse()
        if cleanup:
            self._cleanup(*kw)

    def _cleanup(self, *keywords):
        for key in keywords:
            delattr(self, key)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import io
import openpyxl
import xlrd
import re
import defusedxml
import math
import functools

from ... import Unit


defusedxml.defuse_stdlib()


class SheetReference:

    def _parse_sheet(self, token):
        try:
            sheet, token = token.split('#')
        except ValueError:
            sheet = 0
        else:
            try:
                sheet = int(sheet, 0) - 1
            except (TypeError, ValueError):
                if sheet[0] in ('"', "'") and sheet[~0] == sheet[0] and len(sheet) > 2:
                    sheet = sheet[1:~1]
        return sheet, token

    def _parse_range(self, token):
        try:
            start, end = token.split(':')
            return start, end
        except ValueError:
            return token, token

    def _parse_token(self, token):
        match = re.match(R'^([A-Z]+)(\d+)$', token)
        if not match:
            row, col = (int(x, 0) for x in token.split('.'))
        else:
            col = functools.reduce(
                lambda acc, c: (acc * 26) + c,
                (ord(c) - 0x40 for c in match.group(1)),
                0
            )
            row = int(match.group(2), 10)
        return row, col

    def __init__(self, token):
        self.sheet, token = self._parse_sheet(token)
        start, stop = (self._parse_token(x) for x in self._parse_range(token))
        if start > stop:
            self.stop = start
            self.start = stop
        else:
            self.stop = stop
            self.start = start

    def get_sheet_index(self, names):
        return self.sheet if isinstance(self.sheet, int) else names.index(self.sheet)

    def __iter__(self):
        xmin = min(self.start[0], self.stop[0])
        ymin = min(self.start[1], self.stop[1])
        xmax = max(self.start[0], self.stop[0])
        ymax = max(self.start[1], self.stop[1])
        x, y = xmin, ymin
        while True:
            yield x, y
            if y < ymax:
                y += 1
            elif x < xmax:
                x, y = x + 1, ymin
            else:
                break


class xlxtr(Unit):
    """
    Extract data from Microsoft Excel documents, both Legacy and new XML type
    documents. A sheet reference is of the form `B1` or `1.2`, both specifying
    the first cell of the second column. A cell range can be specified as
    `B1:C12`, or `1.2:C12`, or `1.2:3.12`. Finally, the unit will always refer
    to the first sheet in the document and to change this, specify the sheet
    name or index separated by a hashtag, i.e. `sheet#B1:C12` or `1#B1:C12`.
    Note that indices are 1-based.
    """

    def interface(self, argp):
        argp.add_argument('refs', metavar='reference', type=SheetReference, nargs='*',
            help='A sheet reference to be extracted.')
        argp.epilog = 'If no sheet references are given, the unit lists all sheet names.'
        return super().interface(argp)

    def _process_legacy(self, data):
        import io

        logfile = io.StringIO()
        wb = xlrd.open_workbook(
            file_contents=data,
            logfile=logfile,
            verbosity=self.args.verbose - 1,
            on_demand=True
        )

        logfile.seek(0)
        for entry in logfile:
            entry = entry.strip()
            if re.search(R'^[A-Z]+:', entry) or '***' in entry:
                self.log_info(entry)

        if not self.args.refs:
            self._display_sheet_names(wb.sheet_names())
        for ref in self.args.refs:
            sh = wb.sheet_by_index(ref.get_sheet_index(wb.sheet_names()))
            for cell in ref:
                self.log_debug('emitting cell: ({},{})'.format(*cell))
                col, row = cell
                try:
                    value = sh.cell_value(col - 1, row - 1)
                except IndexError:
                    continue
                if isinstance(value, str):
                    yield value.encode(self.codec)
                else:
                    self.log_warn(F'value not a string in ({col},{row}):', value)

    def _process_xlsx(self, wb):
        if not self.args.refs:
            self._display_sheet_names(wb.sheetnames)
        for ref in self.args.refs:
            sh = wb.worksheets[ref.get_sheet_index(wb.sheetnames)]
            for cell in ref:
                self.log_debug('emitting cell: ({},{})'.format(*cell))
                value = sh.cell(*cell).value
                if value:
                    yield value.encode(self.codec)

    def _display_sheet_names(self, sheets):
        sheets = list(sheets)
        if not sheets:
            return
        width = int(math.log(len(sheets), 10)) + 1
        for k, name in enumerate(sheets):
            self.output(F'sheet {k:{width}d}:', name)

    def process(self, data):
        try:
            workbook = openpyxl.load_workbook(
                io.BytesIO(data),
                read_only=True
            )
        except Exception:
            self.log_info('reverting to xlrd module')
            yield from self._process_legacy(data)
        else:
            yield from self._process_xlsx(workbook)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A package for units that operate primarily on frames of several of inputs.
"""
from .. import Unit
from ...lib.argformats import sliceobj


class FrameSlicer(Unit, abstract=True):

    def interface(self, argp):
        argp.add_argument(
            'slice',
            type=sliceobj,
            nargs='*',
            default=[slice(None, None)],
            help='Specify start:stop:step in Python slice syntax.'
        )
        return super().interface(argp)

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        for s in self.args.slice:
            if s.step and s.step < 0:
                raise ValueError('negative slice steps are not supported here')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import base64
from Crypto.Cipher import AES
from Crypto.Random import urandom
from Crypto.Util.Padding import pad, unpad

from ... import Unit
from ....lib.argformats import multibin, request


class secstr(Unit):
    """
    Implements the AES-based encryption scheme used by the PowerShell commands
    `ConvertFrom-SecureString` and `ConvertTo-SecureString`.
    """

    # This is a magic header value used for PowerShell secure strings.
    _MAGIC = bytes((
        0xEF, 0xAE, 0x3D, 0xD9, 0xDD, 0x75, 0xD7, 0xAE, 0xF8, 0xDD, 0xFD, 0x38,
        0xDB, 0x7E, 0x35, 0xDD, 0xBD, 0x7A, 0xD3, 0x9D, 0x1A, 0xE7, 0x7E, 0x39))
    # Secure strings include a decimal number formatted as a string directly
    # following the header. Presumably, this is the PowerShell version.
    _PSVER = 2
    _IVERR = 'The IV needs to be 16 bytes long.'

    def interface(self, argp):
        argp.add_argument('key', type=multibin, nargs='?', default=bytes(range(1, 17)),
            help='secure string encryption key')
        argp.add_argument('--iv', default=None,
            type=request(multibin, lambda r: len(r) == 0x10, self._IVERR),
            help=F'Optionally specify an IV to use for encryption. {self._IVERR}')
        return super().interface(argp)

    def reverse(self, data):
        ivec = self.args.iv or urandom(0x10)
        if len(ivec) != 0x10:
            raise ValueError(self._IVERR)
        cipher = AES.new(self.args.key, AES.MODE_CBC, ivec)
        data = data.decode('latin-1').encode('utf-16LE')
        data = cipher.encrypt(pad(data, block_size=0x10))
        data = base64.b16encode(data).lower().decode('ascii')
        ivec = base64.b64encode(ivec).decode('ascii')
        data = '|'.join(('%d' % self._PSVER, ivec, data)).encode('utf-16LE')
        return base64.b64encode(self._MAGIC + data)

    def process(self, data):
        head, ivec, data = base64.b64decode(data).split(b'|\0')
        self.log_info('head:', head.hex())
        ivec = base64.b64decode(ivec.decode('utf-16LE'))
        self.log_info('ivec:', ivec.hex())
        data = base64.b16decode(data.decode('utf-16LE'), casefold=True)
        if len(data) % 0x10 != 0:
            self.log_info('data not block-aligned, padding with zeros')
            data += B'\0' * (0x10 - len(data) % 0x10)
        cipher = AES.new(self.args.key, AES.MODE_CBC, ivec)
        data = cipher.decrypt(data)
        try:
            data = unpad(data, block_size=0x10)
        except Exception:
            self.log_warn('decrypted data does not have PKCS7 padding')
        for p in range(0x10):
            try:
                return data[-p:].decode('utf-16LE').encode('latin-1')
            except UnicodeDecodeError:
                pass
            except UnicodeEncodeError:
                pass
        self.log_warn('result is not a padded unicode string, key is likely wrong')
        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import lzma as lzma_

from .. import Unit
from ...lib.argformats import number, OptionFactory, extract_options

__all__ = ['lzma']


class lzma(Unit):
    """
    LZMA compression and decompression.
    """

    def interface(self, argp):
        mode = argp.add_mutually_exclusive_group()
        mode.add_argument('-r', '--raw',
            action='store_true', help='Use raw (no container) format.')
        mode.add_argument('-a', '--alone',
            action='store_true', help='Use the lzma container format.')
        mode.add_argument('-x', '--xz',
            action='store_true', help='Use the default xz format.')
        argp.add_argument('-l', '--level', type=number[0:9], action='store', default=9,
            help='compression level preset between 0 and 9')
        argp.add_argument('-D', '--delta', type=int, default=None, action='store',
            help='Add a delta filter when compressing.')

        filters = extract_options(lzma_, 'FILTER_')
        del filters['DELTA']

        argp.add_argument(
            'filter', nargs='?', type=OptionFactory(filters), choices=list(filters),
            default=None, metavar='FILTER', help=(
                'Specifies a bcj filter to be applied. Possible values '
                'are: {}'.format(', '.join(filters))
            )
        )

        return super().interface(argp)

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.args.level |= lzma_.PRESET_EXTREME

    def _get_lz_mode_and_filters(self, reverse=False):
        mode = lzma_.FORMAT_AUTO
        filters = []
        if self.args.filter is not None:
            filters.append({'id': self.args.filter.value})
        if self.args.delta is not None:
            filters.append({
                'id': lzma_.FILTER_DELTA,
                'dist': self.args.delta
            })
        if self.args.alone:
            mode = lzma_.FORMAT_ALONE
            filters.append({
                'id': lzma_.FILTER_LZMA1,
                'preset': self.args.level
            })
        elif self.args.raw:
            mode = lzma_.FORMAT_RAW
            filters.append({
                'id': lzma_.FILTER_LZMA2,
                'preset': self.args.level
            })
        elif self.args.xz or reverse:
            if reverse:
                self.log_info('choosing default .xz container format for compression.')
            mode = lzma_.FORMAT_XZ
            filters.append({
                'id': lzma_.FILTER_LZMA2,
                'preset': self.args.level
            })
        return mode, filters

    def reverse(self, data):
        mode, filters = self._get_lz_mode_and_filters(True)
        lz = lzma_.LZMACompressor(mode, filters=filters)
        output = lz.compress(data)
        output += lz.flush()
        return output

    def process(self, data):
        keywords = {}
        mode, filters = self._get_lz_mode_and_filters(False)
        if self.args.raw:
            keywords['filters'] = filters
        lz = lzma_.LZMADecompressor(mode, **keywords)
        return lz.decompress(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from fnmatch import fnmatch
from argparse import SUPPRESS

from .... import Unit
from .....lib.dotnet.header import DotNetHeader, ParserEOF
from .....lib.dotnet.resources import NetStructuredResources, NoManagedResource


class dnrc(Unit):
    """
    Extracts all .NET resources whose name matches any of the given patterns
    and outputs them.
    """

    def interface(self, argp):
        argp.add_argument(
            '-s', '--sort', action='store_true', help=(
                'Sort the resourcey by name rather than by order of appearance.'
            )
        )
        argp.add_argument(
            '-r', '--raw', action='count', default=0, help=(
                'Specify once to not deserialize the entries of managed resources. '
                'Use twice to not parse managed resources at all.'
            )
        )
        argp.add_argument(
            'needles', metavar='pattern', nargs='*', type=str, default=[], help=(
                'A wildcard pattern for the name of a .NET resource. '
                'Omit to list all available resources.'
            )
        )
        argp.add_argument('--check', action='store_true', help=SUPPRESS)
        return super().interface(argp)

    def _check(self, major: str, minor: str = ''):
        return any(
            fnmatch(subject, needle) for subject in [
                major, minor, F'{major}.{minor}'
            ]
            for needle in self.args.needles
            if isinstance(needle, str)
        )

    @property
    def _list(self) -> bool:
        return not self.args.needles and not self.args.check

    def process(self, data):
        header = DotNetHeader(data)

        if not header.resources:
            if self._list:
                return
            raise ValueError('This file contains no resources.')

        if self.args.sort:
            header.resources.sort(key=lambda r: r.Name)

        for resource in header.resources:
            try:
                managed = False if self.args.raw > 1 else NetStructuredResources(resource.Data)
            except (NoManagedResource, ParserEOF):
                managed = False

            if not managed:
                if managed is False and self._check(resource.Name):
                    yield resource.Data
                elif self._list:
                    yield resource.Name.encode(self.codec)
                continue

            for entry in managed:
                if entry.Error:
                    self.log_warn(F'entry {resource.Name}.{entry.Name} carried error message: {entry.Error}')
                else:
                    self.log_debug(F'entry {resource.Name}.{entry.Name}')

                if self._list:
                    yield F'{resource.Name}.{entry.Name}'.encode(self.codec)
                    continue
                if not self._check(resource.Name, entry.Name):
                    continue

                if self.args.raw:
                    yield entry.Data
                    continue
                if isinstance(entry.Value, bytes):
                    yield entry.Value
                    continue
                try:
                    yield str(entry.Value).encode(self.codec)
                except Exception:
                    self.log_warn('unable to encode resource value of type', entry.Value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from struct import unpack

from ... import Unit


class carve_zip(Unit):
    """
    Extracts anything from the input data that looks like a zip archive file.
    """

    def process(self, data):

        def carve(start):
            end = start
            while end <= start:
                end = data.find(B'PK\x05\x06', end + 4)
                if end < 0:
                    return None
                size, offset = unpack('<LL', data[end + 12 : end + 20])
                if offset + start + size == end:
                    return end + 22

        cursor = 0

        while True:
            start = data.find(B'PK', cursor)
            if start < cursor:
                break
            end = carve(start)
            if end is None:
                break
            yield data[start:end]
            cursor = end



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from . import PatternExtractor
from ...lib.patterns import formats


class carve(PatternExtractor):
    """
    Extracts patches of data in particular formats from the input.
    """

    def interface(self, argp):
        choices = [p.name for p in formats]
        argp.add_argument(
            'format', type=str, choices=choices, metavar='FORMAT',
            help='Specify one of the following formats: {}'.format(', '.join(choices))
        )
        return super().interface(argp)

    def process(self, data):
        yield from self.matches_processed(data, bytes(formats[self.args.format]))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re

from ...lib.argformats import utf8
from . import RegexUnit, TransformSubstitutionFactory


class resub(RegexUnit):
    """
    A unit for performing substitutions based on a binary regular expression
    pattern. Besides the usual syntax `$k` to insert the `k`-th match group,
    the unit supports processing the contents of match groups with arbitrary
    refinery units (see `refinery.units.pattern.TransformSubstitutionFactory`).
    """

    def interface(self, argp):
        super().interface(argp)
        argp.add_argument('subst', type=utf8, nargs='?', default=B'',
            help='substitution, use $1 for group 1, $0 for entire match. Matches get removed by default.')
        return argp

    def process(self, data):
        self.log_info('pattern:', self.args.regex)
        self.log_info('replace:', self.args.subst)
        return re.sub(self.args.regex, TransformSubstitutionFactory(self.args.subst), data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re
from capstone import Cs, CS_ARCH_X86, CS_ARCH_ARM, CS_MODE_16, CS_MODE_32, CS_MODE_64, CS_MODE_ARM
from string import ascii_letters, digits

from .. import Unit
from ...lib.argformats import number, OptionFactory

__all__ = ['asm']


class box:
    def __init__(self, **kw):
        for k, v in kw.items():
            setattr(self, k, v)


class asm(Unit):
    """
    Disassembles the input data using the capstone disassembly library.
    """

    def interface(self, argp):
        modes = {
            'x16': (CS_ARCH_X86, CS_MODE_16),
            'x32': (CS_ARCH_X86, CS_MODE_32),
            'x64': (CS_ARCH_X86, CS_MODE_64),
            'arm': (CS_ARCH_ARM, CS_MODE_ARM)
        }
        arch = OptionFactory(modes)
        argp.add_argument('mode', choices=list(modes), type=arch, default=arch('x32'), nargs='?', const=arch('x32'),
            help='select architecture for disassembly')
        argp.add_argument('-a', '--addr', action='store_false',
            help='hide addresses of instruction')
        argp.add_argument('-b', '--bytes', action='store_false',
            help='hide instruction bytes next to disassembly')
        argp.add_argument('-s', '--str', action='store_false',
            help='disassemble over detected strings')
        argp.add_argument('-z', '--zeros', action='store_false',
            help='disassemble zero byte patches')
        argp.add_argument('-w', '--width', type=number[3:], default=15,
            help='number of data bytes to put in one row')
        return super().interface(argp)

    def _printable(self, b):
        return 0x20 <= b <= 0x7E and b not in B' \t\v\r\n'

    def _strings(self, data):
        if not self.args.str:
            return
        for match in re.finditer(BR'([ -~]{5,})\x00?|((?:[ -~]\x00){5,})(?:\x00\x00)?', data):
            string = match.group(0)
            try:
                string = string.decode('UTF16-LE') if not string.end \
                    else string.decode('UTF8')
            except Exception:
                continue
            alpha = len([x for x in string if x in digits or x in ascii_letters])
            if 2 * alpha > len(string):
                yield box(start=match.start(), end=match.end(), data=string)
                self.log_info(F'detected string at {match.start():08X}:', string)

    def _format(self, addr=0, data=B'', code='', arg='', comment=''):
        data_str = ''.join('%c' % X if self._printable(X) else '.' for X in data)
        data_hex = ' '.join('%02X' % X for X in data)
        if comment: comment = '    ; ' + comment
        return {
            'addr': addr,
            'str': data_str,
            'hex': data_hex,
            'code': code,
            'arg': arg,
            'comment': comment
        }

    def _bytepatch(self, data, addr, end):
        return self._format(addr, data[addr:end], 'db', ','.join('%02X' % b for b in data[addr:end]))

    def _nullsize(self, data, offset, max):
        length = 0
        try:
            while not data[offset + length] and length < max:
                length += 1
        except IndexError:
            pass
        return length

    def _disassemble(self, data):
        capstone = Cs(*self.args.mode.value)
        strz = self._strings(data)
        string = next(strz, None)
        cursor, done = 0, 0
        while done < len(data):
            cursor = max(cursor, done)
            patchsize = self._nullsize(data, cursor, self.args.width)
            if patchsize > 2:
                yield self._format(done, data[done:done + patchsize], 'db', ','.join('0' * patchsize))
                done += patchsize
                continue
            if cursor >= len(data):
                yield self._bytepatch(data, done, len(data))
                done = cursor
            if string and cursor >= string.end:
                yield self._bytepatch(data, done, string.start)
                yield self._format(string.start, data[string.start:string.end], 'db', string.data)
                done = string.end
                continue
            try:
                ins = next(capstone.disasm(
                    data[cursor:cursor + 15], cursor, count=1))
                end = ins.address + ins.size
                if self.args.str and string:
                    if end > string.start and string.end > cursor:
                        cursor = string.end
                        continue
            except StopIteration:
                cursor += 1
                continue
            else:
                yield self._format(ins.address, ins.bytes, ins.mnemonic, ins.op_str)
                done = end

    def process(self, data):
        disassembly = list(self._disassemble(data))
        for key in ['hex', 'str', 'code', 'arg']:
            m = max(len(r[key]) for r in disassembly)
            for r in disassembly:
                r[key] = r[key].ljust(m)
        line_format = '{code} {arg}{comment}'
        if self.args.bytes:
            line_format = '{hex}  {str}  ' + line_format
        if self.args.addr:
            line_format = '0x{addr:08X}:  ' + line_format
        return '\n'.join(line_format.format(**r) for r in disassembly).encode(self.codec)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import PIL.Image
import io

from .. import Unit


class binpng(Unit):
    """
    Decodes the RGBT (red/green/blue/transparency) values of the pixels
    of a given PNG image file and outputs these values as bytes.
    """
    def process(self, data):
        image = PIL.Image.open(io.BytesIO(data))
        pixelmap, width, height = image.load(), *image.size
        return bytes(code
            for i in range(width)
            for j in range(height)
            for code in pixelmap[i, j])



import refinery as r


class carve_str(r.Unit):
    """
    Carves the longest string expression and removes the surrounding quotes.
    """
    def process(self, data):
        return (r.carve('-lt1', 'string') | r.snip('1:-1'))(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import bz2 as bz2_

from .. import Unit
from ...lib.argformats import number


class bz2(Unit):
    """
    BZip2 compression and decompression.
    """
    def interface(self, argp):
        argp.add_argument('-l', '--level', type=number[1:9], action='store', default=9,
            help='compression level preset between 0 and 9')
        return super().interface(argp)

    def process(self, data):
        return bz2_.decompress(data)

    def reverse(self, data):
        return bz2_.compress(data, self.args.level)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from ...lib.argformats import multibin
from . import BlockTransformation


class transpose(BlockTransformation):
    """
    Interprets the sequence of blocks as rows of a matrix and returns the
    blocks that correspond to the columns of this matrix.
    """
    def interface(self, argp):
        argp.add_argument('padding', type=multibin, default=None, nargs='?',
            help='Optional byte sequence to use as padding for tail end.')
        return super().interface(argp)

    def process(self, data):
        rest = self.rest(data)
        data = list(self.chunk(data, raw=True))

        if self.args.padding:
            while len(rest) < self.args.blocksize:
                rest += self.args.padding
            rest = rest[:self.args.blocksize]
            data.append(rest)
            rest = B''

        return self.unchunk((
            bytes(data[j][i] for j in range(len(data)))
            for i in range(self.args.blocksize)), raw=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Pattern matching based extraction and substitution units.
"""
import re

from typing import Dict, Tuple, Iterable, Optional, Callable, Union, ByteString
from itertools import islice

from ...lib.types import INF, AST
from ...lib.argformats import number, regexp
from .. import Unit


def _lazy_load(cmd: bytes):
    from ...lib.loader import load_commandline
    return load_commandline(cmd.decode('UTF8'))


def TransformSubstitutionFactory(fmt):
    """
    Produces a substitution callable for `refinery.resub` and `refinery.rex`.

    This substitution callable supports refinery operations to be applied to
    match groups. The format `fmt` can contain expressions of the form `$0`,
    `$1`, etc. to be substituted by the corresponding match group from the
    regular expression. Additionally, more complex expressions are supported
    which allow application of arbitrary refinery units to the match group.
    For example, the expression
    ```
    $(1 | hex | add 0x12 | gz)
    ```
    takes the match group with index 1, hex decodes it, adds the value `0x12`
    to every byte and then decompresses it using gzip. To escape a dollar
    symbol, prefix it with an additional dollar symbol.

    Expressions can be nested: For example, `$(1 | snip :$2)` would assume
    that the second match group is an integer k and return only the first k
    bytes of the first match group.
    """

    def unescape_dollars(s):
        return re.sub(RB'(\$+)\$', lambda m: m.group(1), s)

    class PipeLine:
        def __init__(self, argument, seam, pipeline):
            if not argument.isdigit():
                raise ValueError(F'argument "{argument}" is not a digit!')

            def parse_pipeline():
                for command in pipeline:
                    unit = TransformSubstitutionFactory(command)
                    if type(unit) == bytes:
                        unit = _lazy_load(unit)
                    yield unit

            self.argument = int(argument, 10)
            self.seam = unescape_dollars(seam)
            self.units = list(parse_pipeline())

        def __call__(self, match):
            data = match.group(self.argument)
            for unit in self.units:
                try:
                    data = unit(data)
                except AttributeError:
                    unit = _lazy_load(unit(match))
                    data = unit(data)
            return self.seam + data

    try:
        fmt = fmt.decode('UNICODE_ESCAPE')
    except AttributeError:
        if type(fmt) != str:
            raise

    fmt = fmt.encode('UTF8')
    escapes = {m.start() + 1 for m in re.finditer(BR'\$\$', fmt)}
    processors = []
    offset = stop = 0
    pattern = re.compile(BR'\$(\d+|\(\d+)')

    while True:
        expression = pattern.search(fmt, stop)
        if not expression:
            break
        start, stop = expression.span()
        if start in escapes:
            continue
        argument = expression.group(1)
        if not argument.isdigit():
            level = 1
            poles = [expression.start() + 2]
            while level > 0:
                if stop > len(fmt):
                    raise ValueError(F'unbalanced parentheses at {stop} for expression: {fmt}')
                if fmt[stop:stop + 1] == B'(':
                    level += 1
                if fmt[stop:stop + 1] == B')':
                    level -= 1
                if fmt[stop:stop + 1] == B'|' and level == 1:
                    poles.append(stop + 1)
                stop += 1

            poles.append(stop)
            pipeline = [
                fmt[a:b - 1].decode('UTF8').strip()
                for a, b in zip(poles, poles[1:])
            ]

            argument = pipeline.pop(0)
        else:
            pipeline = []

        processors.append(PipeLine(
            argument,
            fmt[offset:start],
            pipeline
        ))
        offset = stop

    epilog = unescape_dollars(fmt[offset:])

    if not offset:
        transformation = epilog
    else:
        def transformation(match):
            return B''.join(p(match) for p in processors) + epilog

    return transformation


class AbstractRegexUnit(Unit, abstract=True):

    def matches(self, data: ByteString, pattern: Union[ByteString, re.Pattern], ascii: bool = True, utf16: bool = True):
        """
        Searches the input data for the given regular expression pattern. If the
        argument `utf16` is `True`, search for occurrences where a zero byte
        is between every character of the match. The `ascii` option allows to
        control whether normal matching results are also returned.
        """
        if not isinstance(pattern, re.Pattern):
            pattern = re.compile(pattern)
        if ascii:
            yield from pattern.finditer(data)
        if utf16:
            d0 = data[0::2]
            d1 = data[1::2]
            for match in pattern.finditer(d0):
                if not any(d1[match.start() + 0 : match.end() + 0]):
                    yield match
            for match in pattern.finditer(d1):
                if not any(d0[match.start() + 1 : match.end() + 1]):
                    yield match


class PatternExtractorBase(AbstractRegexUnit, abstract=True):

    def interface(self, argp):
        argp = super().interface(argp)

        filters = argp.add_argument_group('Match Filters')
        filters.add_argument('-n', '--min', type=number, metavar='N', default=1, help='Matches must have length at least N.')
        filters.add_argument('-N', '--max', type=number, metavar='N', default=INF, help='Matches must have length at most N.')
        filters.add_argument('-E', '--len', type=number, metavar='N', default=AST, help='Matches must be of length N.')
        filters.add_argument('-w', '--whitespace', action='store_true', help='Strip all whitespace from input data.')
        filters.add_argument('-q', '--unique', action='store_true', help='Yield every (transformed) Match only once.')

        filters.set_defaults(ascii=True, utf16=True)

        selectors = argp.add_argument_group('Match Selectors')
        selectors.add_argument('-l', '--longest', action='store_true',
            help='Sort results by length before picking.')

        mode = selectors.add_mutually_exclusive_group()
        mode.add_argument('-t', '--take', type=number[1:], metavar='N', default=INF,
            help='Return the first N occurrences.')
        mode.add_argument('-p', '--pick', type=number, metavar='k', default=None,
            help='Pick the pattern occurrence at index k, starting at 0.')

        return filters

    def matches_filtered(
        self,
        data: ByteString,
        pattern: Union[ByteString, re.Pattern],
        transforms: Optional[Iterable[Union[ByteString, Callable[[re.Match], ByteString]]]] = None,
        early_abort: bool = True
    ) -> Iterable[Tuple[Tuple[int, int], bytes]]:
        """
        This is a wrapper for `AbstractRegexUint.matches` which filters the
        results according to the given commandline arguments. Returns a
        dictionary mapping its position (start, end) in the input data to the
        filtered and transformed match that was found at this position.
        """
        barrier = set()
        taken = 0
        transforms = transforms or [lambda m: m.group(0)]

        if self.args.whitespace:
            data = re.sub(BR'\s+', B'', data)
        for match in self.matches(data, pattern, self.args.ascii, self.args.utf16):
            for transform in transforms:
                hit = transform(match) if callable(transform) else transform
                if hit is None or len(hit) != self.args.len or len(hit) < self.args.min or len(hit) > self.args.max:
                    continue
                if self.args.unique:
                    uid = hash(hit)
                    if uid in barrier:
                        continue
                    barrier.add(uid)
                yield match.span(), hit
                if early_abort:
                    taken += 1
                    if not self.args.longest and taken >= self.args.take:
                        break

    def matches_filtered_cached(
        self,
        data: ByteString,
        pattern: Union[ByteString, re.Pattern],
        transforms: Optional[Iterable[Union[ByteString, Callable[[re.Match], ByteString]]]] = None,
        early_abort: bool = True,
        matches: Optional[dict] = None,
    ) -> Dict[Tuple[int, int], bytes]:
        if matches is None:
            matches = {}
        matches.update(self.matches_filtered(data, pattern, transforms, early_abort))
        return matches

    def matches_finalize(self, matches: Iterable[Tuple[Tuple[int, int], bytes]]) -> Iterable[bytes]:
        """
        Returns the final result of a dictionary of matches according to the
        settings provided via the command line interface.
        """
        result = (d for _, d in matches)
        if self.args.longest:
            result = sorted(result, key=len, reverse=True)
        if self.args.take:
            end = None if self.args.take is INF else self.args.take
            result = islice(result, end)
        if self.args.pick:
            result = islice(result, self.args.pick, self.args.pick + 1)
        yield from result

    def matches_processed(
        self,
        data: ByteString,
        pattern: Union[ByteString, re.Pattern],
        transforms: Optional[Iterable[Union[ByteString, Callable[[re.Match], ByteString]]]] = None
    ) -> Iterable[bytes]:
        """
        A convenience function that acts as the composition of:

        1. `PatternExtractor.matches_filtered` and
        2. `PatternExtractor.matches_finalize`
        """
        yield from self.matches_finalize(self.matches_filtered(data, pattern, transforms))


class PatternExtractor(PatternExtractorBase, abstract=True):
    def interface(self, argp):
        filters = super().interface(argp)
        switch = filters.add_mutually_exclusive_group()
        switch.add_argument(
            '-u', '--only-utf16',
            action='store_false',
            dest='ascii',
            help='Search for UTF16 encoded patterns only.'
        )
        switch.add_argument(
            '-a', '--only-ascii',
            action='store_false',
            dest='utf16',
            help='Search for ASCII encoded patterns only.'
        )
        return argp


class RegexUnit(PatternExtractorBase, abstract=True):

    def interface(self, argp, regex_default=None):
        mode = argp.add_argument_group(
            'Regular Expression Flags',
            'The DOTALL flag is set by default and has to be '
            'disabled rather than enabled.'
        )
        mode.add_argument('-M', '--multiline', action='store_const', default=B'', const=B'm')
        mode.add_argument('-A', '--notdotall', dest='dotall', action='store_const', default=B's', const=B'')
        mode.add_argument('-I', '--ignorecase', action='store_const', default=B'', const=B'i')

        regex = dict(
            type=regexp,
            help='Regular expression to match.'
        )
        if regex_default:
            regex.update(dict(
                default=regex_default,
                nargs='?'
            ))

        argp.add_argument('regex', **regex)

        filters = super().interface(argp)
        filters.add_argument(
            '-u', '--utf16',
            action='store_true',
            default=False,
            help='Search for UTF-16 encoded patterns instead of ASCII.'
        )
        return argp

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.args.ascii = not self.args.utf16
        flags = B''.join(getattr(self.args, flag)
            for flag in ('ignorecase', 'multiline', 'dotall'))
        if flags:
            self.args.regex = B'(?%s)%s' % (flags, self.args.regex)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This package contains all refinery units. To write an executable refinery unit,
it is sufficient to write a class that inherits from `refinery.units.Unit` and
implements `refinery.units.Unit.process`. If the operation implemented by this
unit should be reversible, then a method called `reverse` with the same signature
has to be implemented. For example, the following would be a minimalistic
approach to implement `refinery.hex`:

    from refinery import Unit

    class hex(Unit):
        def process(self, data):
            return bytes.fromhex(data.decode('ascii'))

        def reverse(self, data):
            return data.hex().encode(self.codec)

The above script can be run from the command line. Since `hex` is not marked as
abstract, its inherited `refinery.units.Unit.run` method will be invoked when
the script is executed.
"""
import sys
import os
import io

from enum import IntEnum
from functools import wraps
from inspect import isgenerator
from typing import Iterable, BinaryIO, Union, Optional, Callable, Any, ByteString
from argparse import (
    ArgumentParser,
    ArgumentError,
    RawDescriptionHelpFormatter,
    SUPPRESS,
)

from ..lib.argformats import pending, manifest
from ..lib.tools import terminalfit, get_terminal_size, documentation, lookahead
from ..lib.frame import Framed, Chunk


class ArgparseError(ValueError):
    """
    This custom exception type is thrown from the custom argument parser of
    `refinery.units.Unit` rather than terminating program execution immediately.
    The `parser` parameter is a reference to the argument parser that threw
    the original argument parsing exception with the given `message`.
    """
    def __init__(self, parser, message):
        self.parser = parser
        super().__init__(message)


class RefineryPartialResult(ValueError):
    """
    This exception indicates that a partial result is available.
    """
    def __init__(self, msg, partial):
        super().__init__(msg)
        self.partial = partial


class RefineryCriticalException(RuntimeError):
    """
    If this exception is thrown, processing of the entire input stream
    is aborted instead of just aborting the processing of the current
    chunk.
    """
    pass


class Entry:
    """
    An empty class maker. Any entry point unit (i.e. any unit that can be executed
    via the command line) is an instance of this class.
    """
    pass


class Executable(type):
    """
    This is the metaclass for refinery units. A class which is of this type is
    required to implement a method `run()`. If the class is created in the
    currently executing module, then an instance of the class is automatically
    created after it is defined and its `run()` method is invoked.
    """

    Entry = None
    """
    This variable stores the executable entry point. If more than one entry point
    are present, only the first one is executed and an error message is generated
    for the other ones.
    """

    def __new__(mcs, name, bases, nmspc, abstract=False):
        def normalize(operation: Callable[[Any, ByteString], Any]) -> Callable[[ByteString], Any]:
            @wraps(operation)
            def wrapped(self, data: ByteString) -> bytes:
                if -self.args:
                    if not isinstance(data, bytearray):
                        data = bytearray(data)
                    self.args @= data
                try:
                    return operation(self, data)
                except BaseException as B:
                    return self._exception_handler(B)
            return wrapped

        nmspc.setdefault('__doc__', '')

        for op in ('process', 'reverse'):
            if op in nmspc:
                nmspc[op] = normalize(nmspc[op])

        if not abstract and Entry not in bases:
            bases = bases + (Entry,)
        return super(Executable, mcs).__new__(mcs, name, bases, nmspc)

    def __init__(cls, name, bases, nmspc, abstract=False):
        super(Executable, cls).__init__(name, bases, nmspc)
        if not abstract and sys.modules[cls.__module__].__name__ == '__main__':
            if Executable.Entry:
                cls._output(
                    F'not executing this unit because the following unit was '
                    F'already executed: {Executable.Entry}'
                )
            else:
                Executable.Entry = cls.__name__
                cls.run()

    @property
    def is_reversible(cls) -> bool:
        """
        This property is `True` if and only if the unit has a member function
        named `reverse`. By convention, this member function implements the
        inverse of `refinery.units.Unit.process`.
        """
        return hasattr(cls, 'reverse')

    @property
    def codec(self) -> str:
        """
        The default codec for encoding textual information between units.
        The value of this property is hardcoded to `UTF8`.
        """
        return 'UTF8'


class LogLevel(IntEnum):
    """
    An enumeration representing the current log level:
    """
    DETACHED = -1
    """
    This unit is not attached to a terminal but has been instantiated in
    code. This means that the only way to communicate problems is to throw
    an exception.
    """
    NONE = 0
    """
    Do not log anything.
    """
    WARN = 1
    """
    Default log level: Log warnings.
    """
    INFO = 2
    """
    Increased logging.
    """
    DEBUG = 3
    """
    Maximum logging.
    """


class DelayedArgumentProxy:
    """
    This class implements a proxy for the `args` member variable of `refinery.units.Unit`.
    Its primary purpose is to proxy `refinery.lib.argformats.DelayedArgument` values which
    can be computed only as soon as input data becomes available and which also have to be
    recomputed for each input.
    """

    def __init__(self, argv, argo):
        args = {}
        done = True
        for name, value in vars(argv).items():
            if not pending(value):
                args[name] = value
            else:
                done = False
        self._store(
            _argv=argv,
            _argo=argo,
            _args=args,
            _done=done
        )

    def __neg__(self):
        return not self._done

    def __imatmul__(self, data: bytearray):
        """
        Lock the current arguments for the given input `data`.
        """
        if not self._done:
            for name in self._argo:
                value = getattr(self._argv, name, None)
                if value and pending(value):
                    self._args[name] = manifest(value, data)
        return self

    def _store(self, **kwargs):
        self.__dict__.update(kwargs)

    def __getattr__(self, name):
        try:
            return super().__getattr__(name)
        except AttributeError:
            pass
        try:
            return self._args[name]
        except KeyError:
            pass
        try:
            value = getattr(self._argv, name)
        except AttributeError:
            raise AttributeError(F'Argument {name} not set.')
        if not value or not pending(value):
            return value
        raise AttributeError(F'the value {name} cannot be accessed until data is available.')

    def __setattr__(self, name, value):
        if pending(value):
            self._store(_done=False)
        else:
            self._args[name] = value
        return setattr(self._argv, name, value)


class Unit(metaclass=Executable, abstract=True):
    """
    The base class for all refinery units. It implements a small set of globally
    available options and the handling for multiple inputs and outputs. All units
    implement the _framing_ syntax for producing multiple outputs and ingesting
    multiple inputs in a common format. For more details, see `refinery.lib.frame`.
    """
    @property
    def is_reversible(self) -> bool:
        return self.__class__.is_reversible

    @property
    def codec(self) -> str:
        return self.__class__.codec

    @property
    def log_level(self) -> LogLevel:
        """
        Returns the current log level as an element of `refinery.units.LogLevel`.
        """
        if self.args.quiet:
            return LogLevel.NONE
        return LogLevel(min(len(LogLevel) - 2, self.args.verbose))

    @log_level.setter
    def log_level(self, value: LogLevel) -> None:
        self.args.verbose = int(value)

    def interface(self, argp: ArgumentParser) -> ArgumentParser:
        """
        Receives a reference to an `ArgumentParser` object. This parser will be used to parse
        the command line for this unit into the member variable called `args`.
        Children of `refinery.units.Unit` should override this method to customize their command
        line interface.
        """
        base = argp.add_argument_group(
            'Global Options',
            'The following options are available in every refinery unit.'
        )

        base.add_argument('-h', '--help', action='help',
            help='Show this help message and exit.')
        if self.is_reversible:
            base.add_argument('-R', '--reverse', action='store_true', help='Use the reverse operation.')
        else:
            base.set_defaults(reverse=False)
        base.add_argument('-0', '--null',
            action='store_true', help='Do not produce any output.')

        base.add_argument('-Q', '--quiet', action='store_true', help='Disables all log output.')
        base.add_argument('-L', '--lenient', dest='partial', action='store_true',
            help='Allow this unit to return incomplete results.')
        base.add_argument('-v', '--verbose', action='count',
            help='Specify up to two times to increase log level.')
        base.set_defaults(verbose=1)

        argp.add_argument('--debug-timing', action='store_true', help=SUPPRESS)
        return argp

    def __iter__(self):
        return self

    def _exception_handler(self, exception: BaseException):
        if self.log_level <= LogLevel.DETACHED:
            raise exception
        try:
            raise exception
        except RefineryCriticalException as E:
            self.log_warn(F'critical error, terminating: {E}')
        except RefineryPartialResult as E:
            self.log_warn(F'error, partial result returned: {E}')
            if self.args.partial:
                return E.partial
        except Exception as E:
            self.log_warn(F'unexpected error: {E}')
        finally:
            if self.log_level >= LogLevel.DEBUG:
                import traceback
                traceback.print_exc(file=sys.stderr)

    def __next__(self):
        if not self._chunks:
            self._chunks = iter(self._framehandler)
        while True:
            result = next(self._chunks)
            if not isinstance(result, BaseException):
                return result

    @property
    def _framehandler(self) -> Framed:
        if self._framed:
            return self._framed

        op = self.reverse if self.args.reverse else self.process

        def normalized_action(data: bytearray) -> Iterable[bytes]:
            result = op(data)
            if isgenerator(result):
                yield from filter(lambda x: x is not None, result)
            elif result is not None:
                yield result

        self._framed = Framed(
            normalized_action,
            self.source,
            self.args.nesting,
            self.filter
        )
        return self._framed

    def filter(self, inputs: Iterable[Chunk]) -> Iterable[Chunk]:
        """
        Receives an iterable of `refinery.lib.frame.Chunk`s and yields only those that
        should be processed. The default implementation returns the iterator without
        change; this member function is designed to be overloaded by child classes of
        `refinery.units.Unit` to allow inspection of an entire frame layer and altering
        it before `refinery.units.Unit.process` is called on the individual chunks.
        """
        return inputs

    @property
    def source(self):
        """
        Represents a unit or binary IO stream which has been attached to this unit as its
        source of input data.
        """
        return self._source

    @source.setter
    def source(self, stream):
        if isinstance(stream, self.__class__.__class__):
            stream = stream()
        if not isinstance(stream, self.__class__):
            self._framed = None
            self._chunks = None
        self._source = stream

    @property
    def nozzle(self) -> 'Unit':
        """
        The nozzle is defined recursively as the nozzle of `refinery.units.Unit.source`
        and `self` if no such thing exists. In other words, it is the leftmost unit in
        a pipeline, where data should be inserted for processing.
        """
        try:
            return self.source.nozzle
        except AttributeError:
            return self

    def __class_getitem__(cls, unit: 'Unit'):
        return cls().__getitem__(unit)

    def __getitem__(self, unit: 'Unit'):
        if isinstance(unit, type):
            unit = unit()
        alpha = self.clone()
        omega = unit.clone()
        alpha.args.nesting += 1
        omega.args.nesting -= 1
        omega.nozzle.source = alpha
        return omega

    def __ror__(self, stream: Union[BinaryIO, bytes, bytearray, 'Unit']):
        self.nozzle.source = stream
        return self

    def __or__(self, stream: Union[BinaryIO, 'Unit']):
        try:
            if isinstance(stream, type):
                stream = stream()
            return stream.clone().__ror__(self)
        except AttributeError:
            self._target = stream

        if not self._target.writable():
            return

        def cname(x): return x.lower().replace('-', '')

        if self.isatty and cname(self.codec) != cname(sys.stdout.encoding):
            def recode(chunk):
                import codecs
                return codecs.encode(
                    codecs.decode(chunk, self.codec, errors='backslashreplace'),
                    sys.stdout.encoding
                )
        else:
            def recode(chunk): return chunk

        for last, chunk in lookahead(self):
            if (
                self._framehandler.framebreak
                or self._framehandler.unframed
                and not last
                and not chunk.endswith(B'\n')
            ):
                chunk += B'\n'
            try:
                self._target.write(recode(chunk))
                self._target.flush()
            except AttributeError:
                pass
            except (BrokenPipeError, OSError) as E:
                if isinstance(E, BrokenPipeError) or E.errno != 32:
                    # This happens when the next unit does not consume everything
                    # we send. For example, this can happen when a large file is
                    # read in chunks and the pick unit is used to select only the
                    # first few of these.
                    self.log_info(F'cannot send to next unit: {E}')
                break

        try:
            if self.isatty and not chunk.endswith(B'\n'):
                self._target.write(B'\n')
                self._target.flush()
        except (NameError, AttributeError):
            pass

        return self._target

    def read(self, bytecount: int = -1) -> bytes:
        """
        Reads bytes from the output stream of this unit.
        """
        if not bytecount or bytecount < 0:
            return self.read1()
        bfr = bytearray(bytecount)
        offset = 0
        while offset < bytecount:
            tmp = self.read1(bytecount - offset)
            if not tmp:
                del bfr[offset:]
                break
            end = offset + len(tmp)
            bfr[offset:end] = tmp
            offset = end
        return bytes(bfr)

    def read1(self, bytecount: int = -1) -> bytes:
        """
        Performs a single read against the output stream of this unit and returns
        the result.
        """
        try:
            out = self._buffer or next(self)
            if bytecount and bytecount > 0:
                out, self._buffer = out[:bytecount], out[bytecount:]
            elif self._buffer:
                self._buffer = B''
            return out
        except StopIteration:
            return B''

    def __call__(self, data: Optional[ByteString] = None) -> bytes:
        with io.BytesIO(data) if data else open(os.devnull, 'rb') as stdin:
            with io.BytesIO() as stdout:
                return (stdin | self | stdout).getvalue()

    def process(self, data: bytes) -> Union[Optional[bytes], Iterable[bytes]]:
        """
        This routine is overridden by children of `refinery.units.Unit` to define how
        the unit processes a given chunk of binary data.
        """
        return data

    def log_warn(self, *messages) -> bool:
        """
        Call `refinery.units.Unit.output` for each provided message if and only if the
        current log level is at least `refinery.units.LogLevel.WARN`.
        """
        rv = self.log_level >= LogLevel.WARN
        if rv and messages:
            self.output(*messages)
        return rv

    def log_info(self, *messages) -> bool:
        """
        Call `refinery.units.Unit.output` for each provided message if and only if the
        current log level is at least `refinery.units.LogLevel.INFO`.
        """
        rv = self.log_level >= LogLevel.INFO
        if rv and messages:
            self.output(*messages)
        return rv

    def log_debug(self, *messages) -> bool:
        """
        Call `refinery.units.Unit.output` for each provided message if and only if the
        current log level is at least `refinery.units.LogLevel.DEBUG`.
        """
        rv = self.log_level >= LogLevel.DEBUG
        if rv and messages:
            self.output(*messages)
        return rv

    def output(self, *messages) -> None:
        """
        Logs the provided messages to stderr, prefixed with the current unit's name.
        The routine accepts both string and byte type arguments. Bytestrings are
        decoded with the default codec, using the 'backslashreplace' error handler.
        Does not produce any output if the quiet switch has been enabled via the
        command line arguments.
        """
        if not self.args.quiet:
            return self._output(*messages)

    @property
    def isatty(self) -> bool:
        try:
            return self._target.isatty()
        except AttributeError:
            return False

    @classmethod
    def _output(cls, *messages) -> None:
        def transform(x):
            if isinstance(x, str):
                return x
            if isinstance(x, (bytes, bytearray)):
                import codecs
                return codecs.decode(x, cls.codec, errors='backslashreplace')
            return str(x)
        message = ' '.join(transform(msg) for msg in messages)
        sys.stderr.write(F'{cls.__name__}: {message}\n')

    def clone(self):
        """
        Return a clone of this unit. This routine is used internally to avoid
        mutability issues when using pipe-like syntax for `refinery.units.Unit`
        objects in a Python script.
        """
        other = self.__class__(**vars(self.argv))
        other._source = self._source
        other._buffer = self._buffer
        return other

    def __init__(self, *args, **keywords):
        cols = get_terminal_size()
        args = list(args)

        class ArgumentParserWithKeywordHooks(ArgumentParser):
            def _add_action(self, action):

                class RememberOrder:
                    def __getattr__(self, name):
                        return getattr(action, name)
                    def __setattr__(self, name, value):
                        return setattr(action, name, value)
                    def __call__(self, parser, ns, values, opt=None):
                        if self.dest not in parser.order:
                            parser.order.append(self.dest)
                        return action(parser, ns, values, opt)

                action.required = action.required and action.dest not in keywords
                super()._add_action(RememberOrder())

            def _parse_optional(self, arg_string):
                if isinstance(arg_string, str):
                    return super()._parse_optional(arg_string)

            def error_commandline(self, message):
                super().error(message)

            def error(self, message):
                parser_instance = self
                raise ArgparseError(parser_instance, message)

            def parse_args(self):
                self.order = []
                args_for_parser = args
                if args and args[~0] and isinstance(args[~0], str):
                    nestarg = args[~0]
                    nesting = len(nestarg)
                    if nestarg == ']' * nesting:
                        self.set_defaults(nesting=-nesting)
                        args_for_parser = args[:~0]
                    elif nestarg == '[' * nesting:
                        self.set_defaults(nesting=nesting)
                        args_for_parser = args[:~0]
                self.set_defaults(**keywords)
                try:
                    parsed = super().parse_args(args=args_for_parser)
                except ArgumentError as e:
                    self.error(str(e))
                for name in keywords:
                    param = getattr(parsed, name, None)
                    if param != keywords[name]:
                        self.error(
                            F'parameter "{name}" duplicated with conflicting '
                            F'values {param} and {keywords[name]}'
                        )
                for name in vars(parsed):
                    if name not in self.order:
                        self.order.append(name)
                return parsed

        class LineWrapRawTextHelpFormatter(RawDescriptionHelpFormatter):
            def __init__(self, prog, indent_increment=2, max_help_position=30, width=None):
                super().__init__(prog, indent_increment, max_help_position, width=cols)

            def add_text(self, text):
                if isinstance(text, str):
                    text = terminalfit(text, width=cols)
                return super().add_text(text)

            def _format_action_invocation(self, action):
                if not action.option_strings:
                    metavar, = self._metavar_formatter(action, action.dest)(1)
                    return metavar
                else:
                    parts = []
                    if action.nargs == 0:
                        parts.extend(action.option_strings)
                    else:
                        default = action.dest.upper()
                        args_string = self._format_args(action, default)
                        for option_string in action.option_strings:
                            parts.append(str(option_string))
                        parts[-1] += F' {args_string}'
                    return ', '.join(parts)

        argp = ArgumentParserWithKeywordHooks(
            prog=self.__class__.__name__.replace('_', '-'),
            description=documentation(self.__class__),
            formatter_class=LineWrapRawTextHelpFormatter,
            add_help=False
        )

        argp.set_defaults(nesting=0)

        self.argp = self.interface(argp)
        self.argv = argp.parse_args()
        self.args = DelayedArgumentProxy(self.argv, self.argp.order)

        self._buffer = B''
        self._source = None
        self._target = None
        self._framed = None
        self._chunks = None

        if self.log_level == LogLevel.WARN:
            level = os.environ.get('REFINERY_VERBOSITY', LogLevel.DETACHED)
            try:
                level = int(level)
            except ValueError:
                level = getattr(LogLevel, level, LogLevel.DETACHED)
            self.log_level = level

    @classmethod
    def run(cls, argv=None, stream=None) -> None:
        """
        Implements command line execution. As `refinery.units.Unit` is an `refinery.units.Executable`,
        this method will be executed when a class inheriting from `refinery.units.Unit` is defined in
        the current `__main__` module.
        """
        argv = argv if argv is not None else sys.argv[1:]

        if stream is None:
            stream = open(os.devnull, 'rb') if sys.stdin.isatty() else sys.stdin.buffer

        with stream as source:
            try:
                unit = cls(*argv)
                unit.log_level = max(unit.log_level, LogLevel.WARN)
            except ArgparseError as ap:
                ap.parser.error_commandline(str(ap))
            except Exception as msg:
                cls._output(F'initialization failed:', msg)
            else:
                if unit.args.debug_timing:
                    from time import process_time
                    start_clock = process_time()
                    unit.output('starting clock: {:.4f}'.format(start_clock))

                try:
                    with open(os.devnull, 'wb') if unit.args.null else sys.stdout.buffer as output:
                        source | unit | output
                except KeyboardInterrupt:
                    unit.output('aborting due to keyboard interrupt')
                except OSError:
                    pass

                if unit.args.debug_timing:
                    stop_clock = process_time()
                    unit.output('stopping clock: {:.4f}'.format(stop_clock))
                    unit.output('time delta was: {:.4f}'.format(stop_clock - start_clock))


__pdoc__ = {
    'Unit.is_reversible': Executable.is_reversible.__doc__,
    'Unit.codec': Executable.codec.__doc__
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from .. import Unit
from ...lib.argformats import number


class rep(Unit):
    """
    Duplicates the given input a given number of times.
    """

    def interface(self, argp):
        argp.add_argument('count', type=number[1:], nargs='?', default=2,
            help='Defines the number of outputs for each input. The default is 2.')
        return super().interface(argp)

    def process(self, data: bytes):
        from itertools import repeat
        yield from repeat(data, self.args.count)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from .. import Unit
from ...lib.argformats import number

from .gz import gz
from .ap import aplib
from .zl import zl
from .lz import lzma
from .bz2 import bz2
from .lznt1 import lznt1


class decompress(Unit):
    """
    Attempts all available decompression units against the input and returns
    the output of the first successful one. If none succeeds, the data is
    returned unaltered.
    """
    def interface(self, argp):
        argp.add_argument(
            '-t', '--tolerance',
            metavar='n',
            default=12,
            type=number,
            help='Decompress will attempt to strip up to n bytes from the '
                 'beginning of the data. The default is 12.'
        )
        argp.add_argument(
            '-P', '--no-prepend',
            action='store_false',
            dest='prepend',
            help='By default, if decompression fails, the unit attempts to '
                 'prefix the data with all possible values of a single byte '
                 'and decompress the result. This behavior can be disabled '
                 'with this flag.'
        )
        return super().interface(argp)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.engines = [
            engine() for engine in [zl, lzma, aplib, lznt1, gz, bz2]
        ]

    _MINIMUM_COMPRESSION_FACTOR = 0.5

    def process(self, data):
        for engine in self.engines:
            try:
                result = engine.process(data)
                assert len(result) > len(data) * self._MINIMUM_COMPRESSION_FACTOR
                self.log_info(F'{engine.__class__.__name__} worked.')
                return result
            except Exception:
                pass
        if self.args.prepend:
            for p in range(0x100):
                for engine in self.engines:
                    try:
                        result = engine.process(bytes((p,)) + data)
                        assert len(result) > len(data) * self._MINIMUM_COMPRESSION_FACTOR
                        self.log_info(F'{engine.__class__.__name__} worked after prepending {p:02X}.')
                        return result
                    except Exception:
                        pass
        for t in range(1, self.args.tolerance):
            for engine in self.engines:
                self.log_debug(F'skipping {t:02d}, algorithm {engine.__class__.__name__}')
                try:
                    result = engine.process(data[t:])
                    assert len(result) > len(data) * self._MINIMUM_COMPRESSION_FACTOR
                    self.log_info(F'{engine.__class__.__name__} worked after skipping {t} bytes.')
                    return result
                except Exception:
                    pass
        else:
            self.log_info('nothing worked, returning original data.')
            return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from .. import Unit
from ...lib.argformats import multibin


class ccp(Unit):
    """
    Prepend data to the input.
    """

    def interface(self, argp):
        argp.add_argument('data', nargs='*', type=multibin, help='specify data to be prepended')
        return super().interface(argp)

    def process(self, data):
        return B''.join(self.args.data) + data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re

from .. import Deobfuscator
from ....lib.patterns import formats


class deob_vba_dummy_variables(Deobfuscator):
    def deobfuscate(self, data):
        lines = data.splitlines(keepends=False)
        keeps = {}

        def might_be_used_in(name, line):
            # avoid finding the name within a string literal
            line = '""'.join(re.split(str(formats.ps1str), line))
            line = re.split(RF'\b{name}\b', line)
            try:
                L, R = line
            except ValueError:
                return False
            L = L.strip().lower()
            if L.startswith("'") or L.startswith('rem'):
                return False
            R = R.strip().lower()
            if R.startswith('=') and 'if' not in L:
                return False
            if L.startswith('dim'):
                return False
            return True

        pattern = re.compile(
            R'^\s{0,8}(?:const\s{1,8})?(\w+)\s{1,8}=\s{1,8}.*$',
            flags=re.IGNORECASE
        )

        for k, current_line in enumerate(lines):
            try:
                name = pattern.match(current_line).group(1)
            except AttributeError:
                continue
            used_somewhere = False
            for j, line in enumerate(lines):
                if might_be_used_in(name, line):
                    used_somewhere = keeps[j] = True
            if not used_somewhere and not keeps.get(k, False):
                keeps[k] = False

        return '\n'.join(line for k, line in enumerate(lines) if keeps.get(k, True))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from .. import IterativeDeobfuscator

from .brackets import deob_ps1_brackets
from .concat import deob_ps1_concat
from .escape import deob_ps1_escape
from .cases import deob_ps1_cases
from .format import deob_ps1_format
from .typecast import deob_ps1_typecast
from .stringreplace import deob_ps1_stringreplace
from .literals import deob_ps1_literals
from .invoke import deob_ps1_invoke


class deob_ps1(IterativeDeobfuscator):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sub_units = [sub() for sub in [
            deob_ps1_escape,
            deob_ps1_cases,
            deob_ps1_brackets,
            deob_ps1_format,
            deob_ps1_typecast,
            deob_ps1_stringreplace,
            deob_ps1_concat,
            deob_ps1_invoke,
            deob_ps1_literals
        ]]

    def deobfuscate(self, data):
        for unit in self.sub_units:
            if self.log_debug():
                self.log_debug(F'invoking {unit.__class__.__name__}')
                checkpoint = hash(data)
            data = unit.deobfuscate(data)
            if self.log_debug() and checkpoint != hash(data):
                self.log_debug('data has changed.')
        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Miscellaneous helper functions.
"""
import os
import sys


def format_size(num: int, align=False, default='{} BYTE') -> str:
    """
    Given a number of bytes, produce a human-readable expression for this
    size using common units such as kB and MB.
    """
    step = 1000.0
    result = num
    for unit in ['#B', 'kB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB']:
        if result / step < 0.1:
            break
        result /= step
    if unit == '#B' and not align:
        return default.format(num)
    else:
        width = 6 if align else ''
        return F'{result:{width}.3f} {unit}'


def lookahead(iterator):
    """
    Implements a new iterator from a given one which returns elements
    `(last, item)` where each `item` is taken from the original iterator
    and `last` is a boolean indicating whether this is the last item.
    """
    last = False
    it = iter(iterator)
    try:
        peek = next(it)
    except StopIteration:
        return
    while not last:
        item = peek
        try:
            peek = next(it)
        except StopIteration:
            last = True
        yield last, item


def get_terminal_size():
    """
    Returns the size of the currently attached terminal. If the environment variable
    `REFINERY_TERMSIZE` is set to an integer value, it takes prescedence. If the width
    of the terminal cannot be determined, the function returns zero.
    """
    try:
        return int(os.environ['REFINERY_TERMSIZE'])
    except (KeyError, ValueError):
        pass
    try:
        return os.get_terminal_size(sys.stderr.fileno()).columns - 1
    except OSError:
        return 0


def terminalfit(text: str, delta: int = 0, width: int = 0, **kw) -> str:
    """
    Reformats text to fit the given width while not mangling bullet point lists.
    """
    import textwrap
    width = width or get_terminal_size()
    width = width - delta

    def bulletpoint(line):
        wrapped = textwrap.wrap(line, width - 2, **kw)
        wrapped[1:] = ['  {}'.format(l) for l in wrapped[1:]]
        return '\n'.join(wrapped)

    def fitted(paragraphs):
        for k, p in enumerate(paragraphs):
            if p.startswith(' '):
                yield p
                continue
            if p.startswith('-'):
                input_lines = p.splitlines(keepends=False)
                unwrapped_line = input_lines[0].rstrip()
                lines = []
                if all(t.startswith('-') or t.startswith('  ') for t in input_lines):
                    for line in input_lines[1:]:
                        if not line.startswith('-'):
                            unwrapped_line += ' ' + line.strip()
                            continue
                        lines.append(bulletpoint(unwrapped_line))
                        unwrapped_line = line.rstrip()
                    lines.append(bulletpoint(unwrapped_line))
                    yield '\n'.join(lines)
                    continue
            yield '\n'.join(textwrap.wrap(p, width, **kw))

    return '\n\n'.join(fitted(text.split('\n\n')))


def documentation(unit):
    """
    Return the documentation string of a given unit as it should be displayed
    on the command line. Certain pdoc3-specific reference strings are removed.
    """
    import inspect
    import re
    docs = inspect.getdoc(unit)
    docs = re.sub(R'`refinery\.(?:\w+\.)*(\w+)`', R'\1', docs)
    return docs.replace('`', '')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from . import KeyDerivation


class PasswordDeriveBytes(KeyDerivation):
    """
    An implementation of the PasswordDeriveBytes routine available from the .NET
    standard library. According to documentation, it is an extension of PBKDF1.
    """

    _DEFAULT_SALT = None
    _DEFAULT_HASH = 'SHA1'
    _DEFAULT_ITER = 100

    def process(self, data):
        # TODO move this into a test
        if self.codec != 'UTF8':
            data = data.decode(self.codec).encode('UTF8') + self.salt
        for _ in range(self.iterations - 1):
            data = self.algorithm.new(data).digest()
        seedhash = data
        counter = 1
        data = self.algorithm.new(data).digest()

        while len(data) < self.args.size:
            data += self.algorithm.new(B'%d%s' % (counter, seedhash)).digest()
            counter += 1

        return data[:self.args.size]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Library of regular expression patterns.
"""
import re
import enum

from .tlds import tlds


class pattern:
    """
    A wrapper for regular expression pattern objects created from re.compile,
    allowing combination of several patterns into one via overloaded
    operators.
    """

    def __init__(self, pattern, ignore_case=True):
        self.pattern = pattern
        self.ignore_case = ignore_case
        self.compiled = re.compile(bytes(self), re.IGNORECASE) \
            if ignore_case else re.compile(bytes(self))

    def __bytes__(self):
        return str(self).encode('ascii')

    def __str__(self):
        return self.pattern

    def __getattr__(self, verb):
        return getattr(self.compiled, verb)


class alphabet(pattern):
    """
    A pattern object representing strings of letters from a given alphabet, with
    an optional prefix and postfix.
    """
    def __init__(self, repeat, prefix='', postfix='', at_least=1, at_most=None, **kwargs):
        if not at_most:
            count = '+' if at_least <= 1 else '{{{},}}'.format(at_least)
        else:
            count = '{{{},{}}}(?!{})'.format(at_least, at_most, repeat)

        pattern.__init__(self,
            R'(?<!(?:{r})){b}(?:{r}){c}{a}'.format(
                r=repeat,
                b=prefix,
                c=count,
                a=postfix
            ),
            **kwargs
        )


class tokenize(pattern):
    """
    A pattern representing a sequence of tokens matching the `token` pattern, separated
    by sequences matching the pattern `sep`. The optional parameter `bound` is required
    before and after each token, its default value is the regular expression zero length
    match for a word boundary.
    """
    def __init__(self, token, sep, bound='\\b', **kwargs):
        pattern.__init__(
            self,
            R'(?:{b}{t}{b}{s})+(?:{b}{t}{b})?'.format(
                s=sep, b=bound, t=token),
            **kwargs
        )


class PatternEnum(enum.Enum):
    @classmethod
    def get(cls, name, default):
        try:
            return cls[name]
        except KeyError:
            return default

    def __str__(self):
        return str(self.value)

    def __bytes__(self):
        return bytes(self.value)

    def __repr__(self):
        return F'<pattern {self.name}: {self.value}>'

    def __getattr__(self, name):
        if name in dir(re.Pattern):
            return getattr(self.value, name)
        raise AttributeError


__TLDS = R'(?:{possible_tld})(?!(?:{dealbreakers}))'.format(
    possible_tld='|'.join(tlds),
    dealbreakers='|'.join([
        R'[a-z]',
        R'\.\w\w',
    ])
)

format_domain_normal = R'\b(?:[a-z0-9\_][a-z0-9\-\_]{{0,256}}?\.){repeat}[a-z0-9\_][a-z0-9\-\_]{{1,256}}\.{tlds}'
format_domain_defang = R'\b(?:[a-z0-9\_][a-z0-9\-\_]{{0,256}}?(?:\[\.\]|\.)){repeat}[a-z0-9\_][a-z0-9\-\_]{{1,256}}(?:\[\.\]|\.){tlds}'

pattern_domain = format_domain_normal.format(repeat='{0,20}', tlds=__TLDS)

pattern_subdomain = format_domain_normal.format(repeat='{1,20}', tlds=__TLDS)
pattern_domain_df = format_domain_defang.format(repeat='{0,20}', tlds=__TLDS)

pattern_version = '|'.join('.' * k + 'version' + '.' * (10 - k) for k in range(10))

pattern_ocet = R'(?:[12]\d\d|[1-9]?\d)'
pattern_ipv4 = R'(?<!\.|\d)(?:{o}\.){{3}}{o}(?![\d\.])'.format(o=pattern_ocet)
pattern_ipv4_df = R'(?:{o}(?:\[\.\]|\.)){{2}}{o}\[\.\]{o}'.format(o=pattern_ocet)

pattern_socket = '(?:{ip}|{d})(?::\\d{{2,5}})'.format(ip=pattern_ipv4, d=pattern_domain)
pattern_hostname = pattern_socket + '?'
pattern_hostname_df = '(?:{ip}|{d})(?::\\d{{2,5}})?'.format(ip=pattern_ipv4_df, d=pattern_domain_df)

pattern_integer = '[-+]?(?:0[bB][01]+|0[xX][0-9a-fA-F]+|0[1-7][0-7]*|[1-9][0-9]*|0)(?![a-z0-9])'
pattern_cmdstr = R'''(?:"(?:""|[^"])*"|'(?:''|[^'])*')'''
pattern_ps1str = R'''(?:"(?:\`"|""|[^"])*"|'(?:''|[^'])*')'''
pattern_string = R'''(?:"(?:[^"\\]|\\.)*"|'(?:[^'\\]|\\.)*')'''

pattern_url = ''.join([
    R'([a-z]{2,20}?:\/\/'                     # scheme
    R'(?:[^"\'\s\x00-\x20\x7E-\xFF]{1,256}?'  # username
    R'(?::[^"\'\s\x00-\x20\x7E-\xFF]{0,256}?)?@)?',
    pattern_socket + '?',
    R'(?:[/?#][/_=?&.,\w\%\-]*)?)'
])

pattern_url_df = ''.join([
    R'([a-z]{2,20}?(?:\[:\]|:)\/\/'           # scheme
    R'(?:[^"\'\s\x00-\x20\x7E-\xFF]{1,256}?'  # username
    R'(?::[^"\'\s\x00-\x20\x7E-\xFF]{0,256}?)?@)?',
    pattern_socket + '?',
    R'(?:[/?#][/_=?&.,\w\%\-]*)?)'
])

pattern_email = R'([a-z0-9_\.\+\-]{{1,256}}?)@({})'.format(pattern_domain)
pattern_guid = R'(?:\b|\{)[0-9A-Fa-f]{8}(?:\-[0-9A-Fa-f]{4}){3}\-[0-9A-Fa-f]{12}(?:\}|\b)'

pattern_win_path_nospace = R'[^/\\:"<>|\s\x7E-\xFF\x00-\x1F\xAD]+'
pattern_win_path_element = R'(?:{n} ){{0,4}}{n}'.format(n=pattern_win_path_nospace)
pattern_win_env_variable = R'%[a-zA-Z][a-zA-Z0-9_\-\(\)]{2,}%'
pattern_win_path = R'(?:[A-Za-z]:|{}|\\\\[a-z0-9_.$]+\\[a-z0-9_.$]+)[\\\/]'.format(pattern_win_env_variable)
pattern_win_path += R'(?:{p}[\\\/])*{p}\b'.format(p=pattern_win_path_element)

pattern_hexline = R'(?:{s}+\s+)?\s*{h}(?:\s+{s}+)?'.format(
    h=tokenize(
        R'(?:0x)?[0-9a-f]{2}h?',
        sep=R'[- \t\/:;,\\]{1,3}'
    ).pattern,
    s=R'[-\w:;,#\.\$\?!\/\\=\(\)\[\]\{\}]'
)

pattern_pem = (
    R'-----BEGIN(?:\s[A-Z]+)+-----{n}'
    R'(?:{b}{{40,100}}{n})+{b}{{1,100}}={{0,3}}{n}'
    R'-----END(?:\s[A-Z]+)+-----'
).format(n=R'(?:\r\n|\n\r|\n)', b=R'[0-9a-zA-Z\+\/]')

__all__ = [ 
    'pattern',
    'alphabet',
    'tokenize',
    'formats',
    'indicators',
    'defanged'
]


class formats(PatternEnum):
    """
    An enumeration of patterns for certain formats.
    """
    integer = pattern(pattern_integer)
    "Integer expressions"
    string = pattern(pattern_string)
    "C syntax string literal"
    cmdstr = pattern(pattern_cmdstr)
    "Windows command line escaped string literal"
    ps1str = pattern(pattern_ps1str)
    "PowerShell escaped string literal"
    printable = alphabet(R'[ -~]')
    "Any sequence of printable characters"
    intarray = tokenize(pattern_integer, sep=R'\s*[;,]\s*', bound='')
    "Sequences of integers, separated by commas or semicolons"
    word = alphabet(R'\\w')
    "Sequences of word characters"
    alph = alphabet(R'[a-zA-Z]', ignore_case=False)
    "Sequences of alphabetic characters"
    anum = alphabet(R'[a-zA-Z0-9]', ignore_case=False)
    "Sequences of alpha-numeric characters"
    b64 = alphabet(R'[0-9a-zA-Z\+\/]', postfix=R'[0-9a-zA-Z\+\/]{0,3}={0,3}', ignore_case=False)
    "Base64 encoded strings"
    b64u = alphabet(R'[0-9a-zA-Z\_\-]', postfix=R'[0-9a-zA-Z\_\-]{0,3}={0,3}', ignore_case=False)
    "Base64 encoded strings using URL-safe alphabet"
    hex = alphabet(R'[0-9a-fA-F]')
    "Hexadecimal strings"
    HEX = alphabet(R'[0-9A-F]', ignore_case=False)
    "Uppercase hexadecimal strings"
    hexdump = tokenize(pattern_hexline, bound='', sep=R'\s*\n')
    """
    This pattern matches a typical hexdump output where hexadecimally encoded
    bytes are followed by a string which contains dots or printable characters
    from the dump. For example:

        46 4F 4F 0A 42 41 52 0A  FOO.BAR.
        F0 0B AA BA F0 0B        ......
    """
    hexarray = tokenize(R'[0-9A-F]{2}', sep=R'\s*[;,]\s*', bound='')
    "Arrays of hexadecimal strings, separated by commas or semicolons"


class indicators(PatternEnum):
    """
    An enumeration of patterns for indicators.
    """
    domain = pattern(pattern_domain)
    "Domain names"
    email = pattern(pattern_email)
    "Email addresses"
    guid = pattern(pattern_guid)
    "Windows GUID strings"
    ipv4 = pattern(pattern_ipv4)
    "String representations of IPv4 addresses"
    md5 = alphabet('[0-9a-f]', at_least=32, at_most=32)
    "Hexadecimal strings of length 32"
    sha = alphabet('[0-9a-f]', at_least=40, at_most=40)
    "Hexadecimal strings of length 40"
    sha256 = alphabet('[0-9a-f]', at_least=64, at_most=64)
    "Hexadecimal strings of length 64"
    hostname = pattern(pattern_hostname)
    "Any domain name or IPv4 address, optionally followd by a colon and a port number."
    socket = pattern(pattern_socket)
    "Any domain name or IPv4 address followed by a colon and a (port) number"
    subdomain = pattern(pattern_subdomain)
    "A domain which contains at least three parts, including the top level"
    url = pattern(pattern_url)
    "Uniform resource locator addresses"
    btc = alphabet('[a-km-zA-HJ-NP-Z0-9]', prefix=R'(?<!\w)[13]', at_least=26, at_most=33, ignore_case=False)
    "Bitcoin addresses"
    pem = pattern(pattern_pem, ignore_case=False)
    "A pattern matching PEM encoded cryptographic parameters"
    xmr = alphabet('[1-9A-HJ-NP-Za-km-z]', prefix='4[0-9AB]', at_least=93, at_most=93, ignore_case=False)
    "Monero addresses"
    path = pattern(pattern_win_path)
    "Windows and Linux path names"
    evar = pattern(pattern_win_env_variable)
    "Windows environment variables, i.e. something like `%APPDATA%`"


class defanged(PatternEnum):
    """
    An enumeration of patterns for defanged indicators. Used only by the reverse
    operation of `refinery.defang`.
    """
    hostname = pattern(pattern_hostname_df)
    "A defanged `refinery.lib.patterns.indicators.hostname`."
    url = pattern(pattern_url_df)
    "A defanged `refinery.lib.patterns.indicators.url`."



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import io
from elftools.elf.elffile import ELFFile

from .. import MemoryExtractorUnit


class elfslice(MemoryExtractorUnit):
    """
    Extract data from ELF executables based on virtual offsets.
    """

    def _get_file_offset(self, elf, offset):
        addr = offset.address
        if offset.section:
            for section in elf.iter_sections():
                if section.name == offset.section:
                    addr = section['sh_offset'] + addr
                    return addr, addr + section.size
            else:
                raise ValueError(F'unable to find section {offset.section}.')
        for segment in elf.iter_segments():
            begin = segment.header.p_vaddr
            size = segment.header.p_memsz
            delta = addr - begin
            if delta in range(size + 1):
                return segment.header.p_offset + delta, None
        else:
            raise ValueError('unable to find offset.')

    def _slice(self, off):
        end = off + self.args.limit if self.args.limit else None
        return slice(off, end)

    def process(self, data):
        try:
            elf = ELFFile(io.BytesIO(data))
        except Exception:
            raise ValueError('unable to parse input as ELF file')

        return self._read_from_memory(data,
            lambda addr: self._get_file_offset(elf, addr))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from . import JSONEncoderUnit
from .....lib.dotnet.deserialize import BinaryFormatterParser


class dnds(JSONEncoderUnit):
    """
    Expects data that has been formatted with the .NET class `BinaryFormatter`.
    The output is a representation of the deserialized data in JSON format.
    """

    def interface(self, argp):
        argp.add_argument(
            '-r', '--keep-references',
            dest='dereference',
            action='store_false',
            help='Do not resolve Object references in serialized data.'
        )
        return super().interface(argp)

    def process(self, data):
        self.log_debug('initializing parser, will fail on malformed stream')
        bf = BinaryFormatterParser(
            data,
            keep_meta=True,
            dereference=self.args.dereference,
            ignore_errors=not self.log_debug(),
        )

        return self.to_json([
            {
                'Type': repr(record),
                'Data': record
            } for record in bf
        ])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from .. import Unit
from . import HexViewerMixin


class hexview(Unit, HexViewerMixin):
    """
    Produces a hex dump of the data.
    """

    def interface(self, argp):
        return super().interface(self.hexviewer_interface(argp))

    def process(self, data):
        for line in self.hexdump(data):
            yield line.encode(self.codec)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A collection of compression algorithms. The unit `refinery.decompress`
implements a brute force heuristic decompressor that attempts all known
algorithms against the input data.
"""




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
tlds = ['xn\\-\\-vermgensberatung\\-pwb',
 'xn\\-\\-vermgensberater\\-ctb',
 'xn\\-\\-clchc0ea0b2g2a9gcd',
 'xn\\-\\-w4r85el8fhu5dnra',
 'xn\\-\\-3oq18vl8pn36a',
 'xn\\-\\-5su34j936bgsg',
 'xn\\-\\-bck1b9a5dre4c',
 'xn\\-\\-mgbah1a3hjkrd',
 'xn\\-\\-mgbai9azgqp6j',
 'xn\\-\\-mgberp4a5d4ar',
 'xn\\-\\-xkc2dl3a5ee0h',
 'northwesternmutual',
 'travelersinsurance',
 'xn\\-\\-fzys8d69uvgm',
 'xn\\-\\-mgba7c0bbn0a',
 'xn\\-\\-xkc2al3hye2a',
 'xn\\-\\-i1b6b1a6a2e',
 'xn\\-\\-kcrx77d1x4a',
 'xn\\-\\-lgbbat1ad8j',
 'xn\\-\\-mgba3a4f16a',
 'xn\\-\\-mgbaakc7dvf',
 'xn\\-\\-mgbc0a9azcg',
 'xn\\-\\-nqv7fs00ema',
 'xn\\-\\-54b7fta0cc',
 'xn\\-\\-6qq986b3xl',
 'xn\\-\\-80aqecdr1a',
 'xn\\-\\-b4w605ferd',
 'xn\\-\\-fiq228c5hs',
 'xn\\-\\-h2breg3eve',
 'xn\\-\\-jlq61u9w7b',
 'xn\\-\\-mgba3a3ejt',
 'xn\\-\\-mgbaam7a8h',
 'xn\\-\\-mgbayh7gpa',
 'xn\\-\\-mgbbh1a71e',
 'xn\\-\\-mgbca7dzdo',
 'xn\\-\\-mgbi4ecexp',
 'xn\\-\\-mgbx4cd0ab',
 'xn\\-\\-rvc1e0am3e',
 'americanexpress',
 'kerryproperties',
 'sandvikcoromant',
 'xn\\-\\-eckvdtc9d',
 'xn\\-\\-fpcrj9c3d',
 'xn\\-\\-fzc2c9e2c',
 'xn\\-\\-h2brj9c8c',
 'xn\\-\\-tiq49xqyj',
 'xn\\-\\-yfro4i67o',
 'xn\\-\\-ygbi2ammx',
 'afamilycompany',
 'americanfamily',
 'bananarepublic',
 'cancerresearch',
 'cookingchannel',
 'kerrylogistics',
 'weatherchannel',
 'xn\\-\\-3e0b707e',
 'xn\\-\\-45br5cyl',
 'xn\\-\\-80adxhks',
 'xn\\-\\-80asehdb',
 'xn\\-\\-8y0a063a',
 'xn\\-\\-gckr3f0f',
 'xn\\-\\-mgb9awbf',
 'xn\\-\\-mgbab2bd',
 'xn\\-\\-mgbgu82a',
 'xn\\-\\-mgbpl2fh',
 'xn\\-\\-mgbt3dhd',
 'xn\\-\\-mk1bu44c',
 'xn\\-\\-ngbc5azd',
 'xn\\-\\-ngbe9e0a',
 'xn\\-\\-ogbpf8fl',
 'xn\\-\\-qcka1pmc',
 'international',
 'lifeinsurance',
 'spreadbetting',
 'travelchannel',
 'wolterskluwer',
 'xn\\-\\-11b4c3d',
 'xn\\-\\-1ck2e1b',
 'xn\\-\\-1qqw23a',
 'xn\\-\\-2scrj9c',
 'xn\\-\\-3bst00m',
 'xn\\-\\-3ds443g',
 'xn\\-\\-3hcrj9c',
 'xn\\-\\-42c2d9a',
 'xn\\-\\-45brj9c',
 'xn\\-\\-55qw42g',
 'xn\\-\\-6frz82g',
 'xn\\-\\-80ao21a',
 'xn\\-\\-9krt00a',
 'xn\\-\\-cck2b3b',
 'xn\\-\\-czr694b',
 'xn\\-\\-d1acj3b',
 'xn\\-\\-efvy88h',
 'xn\\-\\-estv75g',
 'xn\\-\\-fct429k',
 'xn\\-\\-fjq720a',
 'xn\\-\\-flw351e',
 'xn\\-\\-g2xx48c',
 'xn\\-\\-gecrj9c',
 'xn\\-\\-gk3at1e',
 'xn\\-\\-h2brj9c',
 'xn\\-\\-hxt814e',
 'xn\\-\\-imr513n',
 'xn\\-\\-j6w193g',
 'xn\\-\\-jvr189m',
 'xn\\-\\-kprw13d',
 'xn\\-\\-kpry57d',
 'xn\\-\\-kpu716f',
 'xn\\-\\-mgbbh1a',
 'xn\\-\\-mgbtx2b',
 'xn\\-\\-mix891f',
 'xn\\-\\-nyqy26a',
 'xn\\-\\-otu796d',
 'xn\\-\\-pbt977c',
 'xn\\-\\-pgbs0dh',
 'xn\\-\\-q9jyb4c',
 'xn\\-\\-rhqv96g',
 'xn\\-\\-rovu88b',
 'xn\\-\\-s9brj9c',
 'xn\\-\\-ses554g',
 'xn\\-\\-t60b56a',
 'xn\\-\\-vuq861b',
 'xn\\-\\-w4rs40l',
 'xn\\-\\-xhq521b',
 'xn\\-\\-zfr164b',
 'construction',
 'lplfinancial',
 'scholarships',
 'versicherung',
 'xn\\-\\-30rr7y',
 'xn\\-\\-3pxu8k',
 'xn\\-\\-45q11c',
 'xn\\-\\-4gbrim',
 'xn\\-\\-55qx5d',
 'xn\\-\\-5tzm5g',
 'xn\\-\\-80aswg',
 'xn\\-\\-90a3ac',
 'xn\\-\\-9dbq2a',
 'xn\\-\\-9et52u',
 'xn\\-\\-c2br7g',
 'xn\\-\\-cg4bki',
 'xn\\-\\-czrs0t',
 'xn\\-\\-czru2d',
 'xn\\-\\-fiq64b',
 'xn\\-\\-fiqs8s',
 'xn\\-\\-fiqz9s',
 'xn\\-\\-io0a7i',
 'xn\\-\\-kput3i',
 'xn\\-\\-mxtq1m',
 'xn\\-\\-o3cw4h',
 'xn\\-\\-pssy2u',
 'xn\\-\\-unup4y',
 'xn\\-\\-wgbh1c',
 'xn\\-\\-wgbl6a',
 'xn\\-\\-y9a3aq',
 'accountants',
 'barclaycard',
 'blackfriday',
 'blockbuster',
 'bridgestone',
 'calvinklein',
 'contractors',
 'creditunion',
 'engineering',
 'enterprises',
 'foodnetwork',
 'investments',
 'kerryhotels',
 'lamborghini',
 'motorcycles',
 'olayangroup',
 'photography',
 'playstation',
 'productions',
 'progressive',
 'redumbrella',
 'rightathome',
 'williamhill',
 'xn\\-\\-90ais',
 'xn\\-\\-c1avg',
 'xn\\-\\-d1alf',
 'xn\\-\\-e1a4c',
 'xn\\-\\-fhbei',
 'xn\\-\\-j1aef',
 'xn\\-\\-j1amh',
 'xn\\-\\-l1acc',
 'xn\\-\\-ngbrx',
 'xn\\-\\-nqv7f',
 'xn\\-\\-p1acf',
 'xn\\-\\-qxa6a',
 'xn\\-\\-tckwe',
 'xn\\-\\-vhquv',
 'accountant',
 'apartments',
 'associates',
 'basketball',
 'bnpparibas',
 'boehringer',
 'capitalone',
 'consulting',
 'creditcard',
 'cuisinella',
 'eurovision',
 'extraspace',
 'foundation',
 'healthcare',
 'immobilien',
 'industries',
 'management',
 'mitsubishi',
 'nationwide',
 'newholland',
 'nextdirect',
 'onyourside',
 'properties',
 'protection',
 'prudential',
 'realestate',
 'republican',
 'restaurant',
 'schaeffler',
 'swiftcover',
 'tatamotors',
 'technology',
 'university',
 'vistaprint',
 'vlaanderen',
 'volkswagen',
 'xn\\-\\-90ae',
 'xn\\-\\-node',
 'xn\\-\\-p1ai',
 'xn\\-\\-qxam',
 'accenture',
 'alfaromeo',
 'allfinanz',
 'amsterdam',
 'analytics',
 'aquarelle',
 'barcelona',
 'bloomberg',
 'christmas',
 'community',
 'directory',
 'education',
 'equipment',
 'fairwinds',
 'financial',
 'firestone',
 'fresenius',
 'frontdoor',
 'fujixerox',
 'furniture',
 'goldpoint',
 'hisamitsu',
 'homedepot',
 'homegoods',
 'homesense',
 'institute',
 'insurance',
 'kuokgroup',
 'lancaster',
 'landrover',
 'lifestyle',
 'marketing',
 'marshalls',
 'melbourne',
 'microsoft',
 'panasonic',
 'passagens',
 'pramerica',
 'richardli',
 'scjohnson',
 'shangrila',
 'solutions',
 'statebank',
 'statefarm',
 'stockholm',
 'travelers',
 'vacations',
 'yodobashi',
 'abudhabi',
 'airforce',
 'allstate',
 'attorney',
 'barclays',
 'barefoot',
 'bargains',
 'baseball',
 'boutique',
 'bradesco',
 'broadway',
 'brussels',
 'budapest',
 'builders',
 'business',
 'capetown',
 'catering',
 'catholic',
 'cipriani',
 'cityeats',
 'cleaning',
 'clinique',
 'clothing',
 'commbank',
 'computer',
 'delivery',
 'deloitte',
 'democrat',
 'diamonds',
 'discount',
 'discover',
 'download',
 'engineer',
 'ericsson',
 'esurance',
 'etisalat',
 'exchange',
 'feedback',
 'fidelity',
 'firmdale',
 'football',
 'frontier',
 'goodyear',
 'grainger',
 'graphics',
 'guardian',
 'hdfcbank',
 'helsinki',
 'holdings',
 'hospital',
 'infiniti',
 'ipiranga',
 'istanbul',
 'jpmorgan',
 'lighting',
 'lundbeck',
 'marriott',
 'maserati',
 'mckinsey',
 'memorial',
 'merckmsd',
 'mortgage',
 'observer',
 'partners',
 'pharmacy',
 'pictures',
 'plumbing',
 'property',
 'redstone',
 'reliance',
 'saarland',
 'samsclub',
 'security',
 'services',
 'shopping',
 'showtime',
 'softbank',
 'software',
 'stcgroup',
 'supplies',
 'symantec',
 'training',
 'vanguard',
 'ventures',
 'verisign',
 'woodside',
 'yokohama',
 'abogado',
 'academy',
 'agakhan',
 'alibaba',
 'android',
 'athleta',
 'auction',
 'audible',
 'auspost',
 'avianca',
 'banamex',
 'bauhaus',
 'bentley',
 'bestbuy',
 'booking',
 'brother',
 'bugatti',
 'capital',
 'caravan',
 'careers',
 'channel',
 'charity',
 'chintai',
 'citadel',
 'clubmed',
 'college',
 'cologne',
 'comcast',
 'company',
 'compare',
 'contact',
 'cooking',
 'corsica',
 'country',
 'coupons',
 'courses',
 'cricket',
 'cruises',
 'dentist',
 'digital',
 'domains',
 'exposed',
 'express',
 'farmers',
 'fashion',
 'ferrari',
 'ferrero',
 'finance',
 'fishing',
 'fitness',
 'flights',
 'florist',
 'flowers',
 'forsale',
 'frogans',
 'fujitsu',
 'gallery',
 'genting',
 'godaddy',
 'grocery',
 'guitars',
 'hamburg',
 'hangout',
 'hitachi',
 'holiday',
 'hosting',
 'hoteles',
 'hotmail',
 'hyundai',
 'ismaili',
 'jewelry',
 'juniper',
 'kitchen',
 'komatsu',
 'lacaixa',
 'lanxess',
 'lasalle',
 'latrobe',
 'leclerc',
 'limited',
 'lincoln',
 'markets',
 'metlife',
 'monster',
 'netbank',
 'netflix',
 'network',
 'neustar',
 'okinawa',
 'oldnavy',
 'organic',
 'origins',
 'philips',
 'pioneer',
 'politie',
 'realtor',
 'recipes',
 'rentals',
 'reviews',
 'rexroth',
 'samsung',
 'sandvik',
 'schmidt',
 'schwarz',
 'science',
 'shiksha',
 'shriram',
 'singles',
 'staples',
 'storage',
 'support',
 'surgery',
 'systems',
 'temasek',
 'theater',
 'theatre',
 'tickets',
 'tiffany',
 'toshiba',
 'trading',
 'walmart',
 'wanggou',
 'watches',
 'weather',
 'website',
 'wedding',
 'whoswho',
 'windows',
 'winners',
 'xfinity',
 'yamaxun',
 'youtube',
 'zuerich',
 'abarth',
 'abbott',
 'abbvie',
 'africa',
 'agency',
 'airbus',
 'airtel',
 'alipay',
 'alsace',
 'alstom',
 'anquan',
 'aramco',
 'author',
 'bayern',
 'beauty',
 'berlin',
 'bharti',
 'bostik',
 'boston',
 'broker',
 'camera',
 'career',
 'caseih',
 'casino',
 'center',
 'chanel',
 'chrome',
 'church',
 'circle',
 'claims',
 'clinic',
 'coffee',
 'comsec',
 'condos',
 'coupon',
 'credit',
 'cruise',
 'dating',
 'datsun',
 'dealer',
 'degree',
 'dental',
 'design',
 'direct',
 'doctor',
 'dunlop',
 'dupont',
 'durban',
 'emerck',
 'energy',
 'estate',
 'events',
 'expert',
 'family',
 'flickr',
 'futbol',
 'gallup',
 'garden',
 'george',
 'giving',
 'global',
 'google',
 'gratis',
 'health',
 'hermes',
 'hiphop',
 'hockey',
 'hotels',
 'hughes',
 'imamat',
 'insure',
 'intuit',
 'jaguar',
 'joburg',
 'juegos',
 'kaufen',
 'kinder',
 'kindle',
 'kosher',
 'lancia',
 'latino',
 'lawyer',
 'lefrak',
 'living',
 'locker',
 'london',
 'luxury',
 'madrid',
 'maison',
 'makeup',
 'market',
 'mattel',
 'mobile',
 'monash',
 'mormon',
 'moscow',
 'museum',
 'mutual',
 'nagoya',
 'natura',
 'nissan',
 'nissay',
 'norton',
 'nowruz',
 'office',
 'olayan',
 'online',
 'oracle',
 'orange',
 'otsuka',
 'pfizer',
 'photos',
 'physio',
 'pictet',
 'quebec',
 'racing',
 'realty',
 'reisen',
 'repair',
 'report',
 'review',
 'rocher',
 'rogers',
 'ryukyu',
 'safety',
 'sakura',
 'sanofi',
 'school',
 'schule',
 'search',
 'secure',
 'select',
 'shouji',
 'soccer',
 'social',
 'stream',
 'studio',
 'supply',
 'suzuki',
 'swatch',
 'sydney',
 'taipei',
 'taobao',
 'target',
 'tattoo',
 'tennis',
 'tienda',
 'tjmaxx',
 'tkmaxx',
 'toyota',
 'travel',
 'unicom',
 'viajes',
 'viking',
 'villas',
 'virgin',
 'vision',
 'voting',
 'voyage',
 'vuelos',
 'walter',
 'webcam',
 'xihuan',
 'yachts',
 'yandex',
 'zappos',
 'actor',
 'adult',
 'aetna',
 'amfam',
 'amica',
 'apple',
 'archi',
 'audio',
 'autos',
 'azure',
 'baidu',
 'beats',
 'bible',
 'bingo',
 'black',
 'boats',
 'bosch',
 'build',
 'canon',
 'cards',
 'chase',
 'cheap',
 'cisco',
 'citic',
 'click',
 'cloud',
 'coach',
 'codes',
 'crown',
 'cymru',
 'dabur',
 'dance',
 'deals',
 'delta',
 'drive',
 'dubai',
 'earth',
 'edeka',
 'email',
 'epson',
 'faith',
 'fedex',
 'final',
 'forex',
 'forum',
 'gallo',
 'games',
 'gifts',
 'gives',
 'glade',
 'glass',
 'globo',
 'gmail',
 'green',
 'gripe',
 'group',
 'gucci',
 'guide',
 'homes',
 'honda',
 'horse',
 'house',
 'hyatt',
 'ikano',
 'intel',
 'irish',
 'iveco',
 'jetzt',
 'koeln',
 'kyoto',
 'lamer',
 'lease',
 'legal',
 'lexus',
 'lilly',
 'linde',
 'lipsy',
 'lixil',
 'loans',
 'locus',
 'lotte',
 'lotto',
 'lupin',
 'macys',
 'mango',
 'media',
 'miami',
 'money',
 'movie',
 'nadex',
 'nexus',
 'nikon',
 'ninja',
 'nokia',
 'nowtv',
 'omega',
 'onion',
 'osaka',
 'paris',
 'parts',
 'party',
 'phone',
 'photo',
 'pizza',
 'place',
 'poker',
 'praxi',
 'press',
 'prime',
 'promo',
 'quest',
 'radio',
 'rehab',
 'reise',
 'ricoh',
 'rocks',
 'rodeo',
 'rugby',
 'salon',
 'sener',
 'seven',
 'sharp',
 'shell',
 'shoes',
 'skype',
 'sling',
 'smart',
 'smile',
 'solar',
 'space',
 'sport',
 'stada',
 'store',
 'study',
 'style',
 'sucks',
 'swiss',
 'tatar',
 'tires',
 'tirol',
 'tmall',
 'today',
 'tokyo',
 'tools',
 'toray',
 'total',
 'tours',
 'trade',
 'trust',
 'tunes',
 'tushu',
 'ubank',
 'vegas',
 'video',
 'vodka',
 'volvo',
 'wales',
 'watch',
 'weber',
 'weibo',
 'works',
 'world',
 'xerox',
 'yahoo',
 'aarp',
 'able',
 'adac',
 'aero',
 'aigo',
 'akdn',
 'ally',
 'amex',
 'arab',
 'army',
 'arpa',
 'arte',
 'asda',
 'asia',
 'audi',
 'auto',
 'baby',
 'band',
 'bank',
 'bbva',
 'beer',
 'best',
 'bike',
 'bing',
 'blog',
 'blue',
 'bofa',
 'bond',
 'book',
 'buzz',
 'cafe',
 'call',
 'camp',
 'care',
 'cars',
 'casa',
 'case',
 'cash',
 'cbre',
 'cern',
 'chat',
 'citi',
 'city',
 'club',
 'cool',
 'coop',
 'cyou',
 'data',
 'date',
 'dclk',
 'deal',
 'dell',
 'desi',
 'diet',
 'dish',
 'docs',
 'duck',
 'dvag',
 'erni',
 'fage',
 'fail',
 'fans',
 'farm',
 'fast',
 'fiat',
 'fido',
 'film',
 'fire',
 'fish',
 'flir',
 'food',
 'ford',
 'free',
 'fund',
 'game',
 'gbiz',
 'gent',
 'ggee',
 'gift',
 'gmbh',
 'gold',
 'golf',
 'goog',
 'guge',
 'guru',
 'hair',
 'haus',
 'hdfc',
 'help',
 'here',
 'hgtv',
 'host',
 'hsbc',
 'icbc',
 'ieee',
 'imdb',
 'immo',
 'info',
 'itau',
 'java',
 'jeep',
 'jobs',
 'jprs',
 'kddi',
 'kiwi',
 'kpmg',
 'kred',
 'land',
 'lego',
 'lgbt',
 'lidl',
 'life',
 'like',
 'limo',
 'link',
 'live',
 'loan',
 'loft',
 'love',
 'ltda',
 'luxe',
 'maif',
 'meet',
 'meme',
 'menu',
 'mini',
 'mint',
 'mobi',
 'moda',
 'moto',
 'name',
 'navy',
 'news',
 'next',
 'nico',
 'nike',
 'ollo',
 'open',
 'page',
 'pars',
 'pccw',
 'pics',
 'ping',
 'pink',
 'play',
 'plus',
 'pohl',
 'porn',
 'post',
 'prod',
 'prof',
 'qpon',
 'raid',
 'read',
 'reit',
 'rent',
 'rest',
 'rich',
 'rmit',
 'room',
 'rsvp',
 'ruhr',
 'safe',
 'sale',
 'sarl',
 'save',
 'saxo',
 'scor',
 'scot',
 'seat',
 'seek',
 'sexy',
 'shaw',
 'shia',
 'shop',
 'show',
 'silk',
 'sina',
 'site',
 'skin',
 'sncf',
 'sohu',
 'song',
 'sony',
 'spot',
 'star',
 'surf',
 'talk',
 'taxi',
 'team',
 'tech',
 'teva',
 'tiaa',
 'tips',
 'town',
 'toys',
 'tube',
 'vana',
 'visa',
 'viva',
 'vivo',
 'vote',
 'voto',
 'wang',
 'weir',
 'wien',
 'wiki',
 'wine',
 'work',
 'xbox',
 'yoga',
 'zara',
 'zero',
 'zone',
 'aaa',
 'abb',
 'abc',
 'aco',
 'ads',
 'aeg',
 'afl',
 'aig',
 'anz',
 'aol',
 'app',
 'art',
 'aws',
 'axa',
 'bar',
 'bbc',
 'bbt',
 'bcg',
 'bcn',
 'bet',
 'bid',
 'bio',
 'bit',
 'biz',
 'bms',
 'bmw',
 'bom',
 'boo',
 'bot',
 'box',
 'buy',
 'bzh',
 'cab',
 'cal',
 'cam',
 'car',
 'cat',
 'cba',
 'cbn',
 'cbs',
 'ceb',
 'ceo',
 'cfa',
 'cfd',
 'com',
 'cpa',
 'crs',
 'csc',
 'dad',
 'day',
 'dds',
 'dev',
 'dhl',
 'diy',
 'dnp',
 'dog',
 'dot',
 'dtv',
 'dvr',
 'eat',
 'eco',
 'edu',
 'esq',
 'eus',
 'fan',
 'fit',
 'fly',
 'foo',
 'fox',
 'frl',
 'ftr',
 'fun',
 'fyi',
 'gal',
 'gap',
 'gay',
 'gdn',
 'gea',
 'gle',
 'gmo',
 'gmx',
 'goo',
 'gop',
 'got',
 'gov',
 'hbo',
 'hiv',
 'hkt',
 'hot',
 'how',
 'ibm',
 'ice',
 'icu',
 'ifm',
 'inc',
 'ing',
 'ink',
 'int',
 'ist',
 'itv',
 'jcb',
 'jcp',
 'jio',
 'jll',
 'jmp',
 'jnj',
 'jot',
 'joy',
 'kfh',
 'kia',
 'kim',
 'kpn',
 'krd',
 'lat',
 'law',
 'lds',
 'llc',
 'llp',
 'lol',
 'lpl',
 'ltd',
 'man',
 'map',
 'mba',
 'med',
 'men',
 'mil',
 'mit',
 'mlb',
 'mls',
 'mma',
 'moe',
 'moi',
 'mom',
 'mov',
 'msd',
 'mtn',
 'mtr',
 'nab',
 'nba',
 'nec',
 'net',
 'new',
 'nfl',
 'ngo',
 'nhk',
 'now',
 'nra',
 'nrw',
 'ntt',
 'nyc',
 'obi',
 'off',
 'one',
 'ong',
 'onl',
 'ooo',
 'org',
 'ott',
 'ovh',
 'pay',
 'pet',
 'phd',
 'pid',
 'pin',
 'pnc',
 'pro',
 'pru',
 'pub',
 'pwc',
 'qvc',
 'red',
 'ren',
 'ril',
 'rio',
 'rip',
 'run',
 'rwe',
 'sap',
 'sas',
 'sbi',
 'sbs',
 'sca',
 'scb',
 'ses',
 'sew',
 'sex',
 'sfr',
 'ski',
 'sky',
 'soy',
 'srl',
 'stc',
 'tab',
 'tax',
 'tci',
 'tdk',
 'tel',
 'thd',
 'tjx',
 'top',
 'trv',
 'tui',
 'tvs',
 'ubs',
 'uno',
 'uol',
 'ups',
 'vet',
 'vig',
 'vin',
 'vip',
 'wed',
 'win',
 'wme',
 'wow',
 'wtc',
 'wtf',
 'xin',
 'xxx',
 'xyz',
 'you',
 'yun',
 'zip',
 'ac',
 'ad',
 'ae',
 'af',
 'ag',
 'ai',
 'al',
 'am',
 'ao',
 'aq',
 'ar',
 'as',
 'at',
 'au',
 'aw',
 'ax',
 'az',
 'ba',
 'bb',
 'bd',
 'be',
 'bf',
 'bg',
 'bh',
 'bi',
 'bj',
 'bm',
 'bn',
 'bo',
 'br',
 'bs',
 'bt',
 'bv',
 'bw',
 'by',
 'bz',
 'ca',
 'cc',
 'cd',
 'cf',
 'cg',
 'ch',
 'ci',
 'ck',
 'cl',
 'cm',
 'cn',
 'co',
 'cr',
 'cu',
 'cv',
 'cw',
 'cx',
 'cy',
 'cz',
 'de',
 'dj',
 'dk',
 'dm',
 'do',
 'dz',
 'ec',
 'ee',
 'eg',
 'er',
 'es',
 'et',
 'eu',
 'fi',
 'fj',
 'fk',
 'fm',
 'fo',
 'fr',
 'ga',
 'gb',
 'gd',
 'ge',
 'gf',
 'gg',
 'gh',
 'gi',
 'gl',
 'gm',
 'gn',
 'gp',
 'gq',
 'gr',
 'gs',
 'gt',
 'gu',
 'gw',
 'gy',
 'hk',
 'hm',
 'hn',
 'hr',
 'ht',
 'hu',
 'id',
 'ie',
 'il',
 'im',
 'in',
 'io',
 'iq',
 'ir',
 'is',
 'it',
 'je',
 'jm',
 'jo',
 'jp',
 'ke',
 'kg',
 'kh',
 'ki',
 'km',
 'kn',
 'kp',
 'kr',
 'kw',
 'ky',
 'kz',
 'la',
 'lb',
 'lc',
 'li',
 'lk',
 'lr',
 'ls',
 'lt',
 'lu',
 'lv',
 'ly',
 'ma',
 'mc',
 'md',
 'me',
 'mg',
 'mh',
 'mk',
 'ml',
 'mm',
 'mn',
 'mo',
 'mp',
 'mq',
 'mr',
 'ms',
 'mt',
 'mu',
 'mv',
 'mw',
 'mx',
 'my',
 'mz',
 'na',
 'nc',
 'ne',
 'nf',
 'ng',
 'ni',
 'nl',
 'no',
 'np',
 'nr',
 'nu',
 'nz',
 'om',
 'pa',
 'pe',
 'pf',
 'pg',
 'ph',
 'pk',
 'pl',
 'pm',
 'pn',
 'pr',
 'ps',
 'pt',
 'pw',
 'py',
 'qa',
 're',
 'ro',
 'rs',
 'ru',
 'rw',
 'sa',
 'sb',
 'sc',
 'sd',
 'se',
 'sg',
 'sh',
 'si',
 'sj',
 'sk',
 'sl',
 'sm',
 'sn',
 'so',
 'sr',
 'ss',
 'st',
 'su',
 'sv',
 'sx',
 'sy',
 'sz',
 'tc',
 'td',
 'tf',
 'tg',
 'th',
 'tj',
 'tk',
 'tl',
 'tm',
 'tn',
 'to',
 'tr',
 'tt',
 'tv',
 'tw',
 'tz',
 'ua',
 'ug',
 'uk',
 'us',
 'uy',
 'uz',
 'va',
 'vc',
 've',
 'vg',
 'vi',
 'vn',
 'vu',
 'wf',
 'ws',
 'ye',
 'yt',
 'za',
 'zm',
 'zw']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re

from . import BlockTransformation
from ...lib.argformats import number
from ...lib.patterns import formats
from ..encoding.base import base as base_unit


class pack(BlockTransformation):
    """
    Scans the input data for numeric constants and packs them into a binary
    format. This is useful to convert the textual representation of an array of
    numbers into its binary form. For example, `123,34,256,12,1,234` would be
    transformed into the byte sequence `7B22000C01EA`, where `256` was wrapped
    and packed as a null byte because the default block size is one byte. If
    the above sequence would be packed with options -EB2, the result would be
    equal to `007B00220100000C000100EA` in hexadecimal.
    """

    def interface(self, argp):
        base = argp.add_argument_group(
            'Number Base Options',
            'Up to base 36 is supported by extending the decimal digits with '
            'case-insensitive alphabetic characters A through Z.'
        ).add_mutually_exclusive_group()
        base.add_argument(
            'base',
            type=number[2:36],
            default=0,
            nargs='?',
            help='Find only numbers in given base. Default of 0 means that '
                 'common expressions for hexadecimal, octal and binary are '
                 'accepted.')
        base.add_argument(
            '-x', '--hexdump',
            action='store_true',
            help='Only look for exactly two digit hexadecimal numbers surrounded by whitespace.'
        )
        rev = argp.add_argument_group(
            'Reverse Options',
            'The following options only apply to reverse mode.'
        )
        rev.add_argument('-P', '--no-prefix', action='store_true',
            help='Does not automatically add numeric prefixes in reverse mode')
        return super().interface(argp)

    @property
    def bytestream(self):
        # never alow bytes to be left unchunked
        return False

    def reverse(self, data):
        base = self.args.base or 10
        prefix = B''

        self.log_debug(F'using base {base:d}')

        if not self.args.no_prefix:
            prefix = {
                0x02: b'0b',
                0x08: b'0o',
                0x10: b'0x'
            }.get(base, B'')

        converter = base_unit(
            base=base,
            byteorder=('big', 'little')[self.args.little_endian]
        )

        for n in self.chunk(data, raw=True):
            yield prefix + converter.reverse(n)

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        if self.args.hexdump:
            self.log_debug('enabling hexdump parser')
            self.args.base = 0x10

    def process(self, data):
        if self.args.hexdump:
            pattern = re.compile(BR'(?:\s|^)(?:0x)?([0-9a-f]{%i})h?(?=\s|$)' % (self.args.blocksize * 2), re.IGNORECASE)
        elif self.args.base == 0:
            pattern = formats.integer
        elif self.args.base <= 10:
            pattern = re.compile(B'[-+]?[0-%d]{1,64}' % (self.args.base - 1))
        else:
            pattern = re.compile(B'[-+]?[0-9a-%c]{1,20}' % (0x57 + self.args.base), re.IGNORECASE)

        items = pattern.findall(data)
        items = [int(n, self.args.base) & self.fmask for n in items]
        return self.unchunk(items)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re

from .. import IterativeDeobfuscator
from . import string_unquote, string_quote
from ....lib.patterns import formats


class deob_ps1_concat(IterativeDeobfuscator):
    def deobfuscate(self, data):
        def concatenate(match):
            a, b = match.groups()
            return string_quote(string_unquote(a) + string_unquote(b))
        return re.sub(
            R'({s})\s*[\+\&]{{1}}\s*({s})'.format(s=formats.ps1str),
            concatenate,
            data
        )



# -*- coding: utf-8 -*-
"""
Parsing of managed .NET resources, which are .NET resource directories which
begin with the magic sequence `0xBEEFCACE`. These resources can contain several
entries of serialized data. The main reference used for this parser was the
dnSpy source code.
"""
import re

from .deserialize import BinaryFormatterParser
from .types import (
    Blob,
    Box,
    Byte,
    Char,
    LengthPrefixedString,
    StreamReader,
    StringPrimitive,
    EncodedInteger,
    Struct,
    UInt16,
    UInt32,
    Int16,
    Int32,
    Int64,
    SByte,
    Single,
    Double,
    Null,
    UInt64,
    unpack,
    DateTime,
    TimeSpan
)


class NoManagedResource(AssertionError):
    pass


class String(LengthPrefixedString):
    def __init__(self, reader):
        LengthPrefixedString.__init__(self, reader, codec='UTF-8')


class Boolean(Byte):
    @property
    def Value(self):
        return bool(super(Boolean, self).Value)


class Decimal(Blob):
    def __init__(self, reader):
        Blob.__init__(self, reader, 16)

    @property
    def Value(self):
        # TODO: Unknown whether this is correct
        return int.from_bytes(self._data, 'big')


class ByteArray(Struct):
    def parse(self):
        self.Size = self.expect(UInt32)
        self.Value = self._reader.read(self.Size)

    def __bytes__(self):
        return self.Value


class NetManifestResource(Struct):
    USERTYPES = 0x40
    PRIMITIVE = {
        0x00: Null,
        0x01: String,
        0x02: Boolean,
        0x03: Char,
        0x04: Byte,
        0x05: SByte,
        0x06: Int16,
        0x07: UInt16,
        0x08: Int32,
        0x09: UInt32,
        0x0A: Int64,
        0x0B: UInt64,
        0x0C: Single,
        0x0D: Double,
        0x0E: Decimal,
        0x0F: DateTime,
        0x10: TimeSpan,
        0x20: ByteArray,
        0x21: ByteArray,
    }

    def parse(self):
        self.Signature = self.expect(UInt32)
        if self.Signature != 0xBEEFCACE:
            raise NoManagedResource

        self.ReaderCount = self.expect(UInt32)
        self.ReaderTypeLength = self.expect(UInt32)

        tr = StreamReader(self._reader.read(self.ReaderTypeLength))
        self.ReaderType = tr.expect(StringPrimitive)
        self.ResourceSetType = tr.expect(StringPrimitive)

        if not re.match(r"^System\.Resources\.ResourceReader,\s*mscorlib", self.ReaderType):
            raise AssertionError('unknown resource reader')

        self.Version = self.expect(UInt32)
        ResourceCount = self.expect(UInt32)
        RsrcTypeCount = self.expect(UInt32)

        ResourceTypes = [
            self.expect(StringPrimitive)
            for _ in range(RsrcTypeCount)
        ]

        self._reader.align(8)
        self._reader.skip(4 * ResourceCount)

        # Since we do not require the resouce hashes, we skip over them.
        # The following would be the code to read in the hashes:
        #
        # ResourceHashes = [
        #     self.expect(UInt32)
        #     for _ in range(ResourceCount)
        # ]

        ResourceNameOffsets = [
            self.expect(UInt32)
            for _ in range(ResourceCount)
        ]

        self.DataSectionOffset = self.expect(UInt32)

        self.Resources = []

        for k in range(ResourceCount):
            with self._reader.checkpoint():
                self._reader.skip(ResourceNameOffsets[k])
                Name = self.expect(LengthPrefixedString, codec='UTF-16LE')
                Offset = self.expect(UInt32) + self.DataSectionOffset
                self.Resources.append(Box(Offset=Offset, Name=Name))

        self.Resources.sort(key=lambda r: r.Offset)
        self.Resources.append(Box(Offset=len(self._reader)))
        self.Resources = [
            Box(Size=b.Offset - a.Offset - 1, **a)
            for a, b in zip(self.Resources, self.Resources[1:])
        ]

        for Index, Entry in enumerate(self.Resources):

            self._reader.seek(Entry.Offset)
            TypeCode = self.expect(EncodedInteger)
            Entry.Error = None
            Entry.Value = Entry.Data = self._reader.read(Entry.Size)

            if TypeCode >= self.USERTYPES:
                Entry.TypeName = ResourceTypes[TypeCode - self.USERTYPES]
                try:
                    Deserialized = BinaryFormatterParser(
                        Entry.Data,
                        ignore_errors=False,
                        dereference=False,
                        keep_meta=False
                    )
                except Exception as error:
                    Entry.Error = 'failed to deserialize entry data: {}'.format(error)
                    continue
                try:
                    _, _, _, Data = Deserialized
                except ValueError:
                    Entry.Error = 'deserialized entry has {} records, 4 were expected.'.format(len(Deserialized))
                    continue
                if Data not in Entry.Data:
                    Entry.Error = 'the computed entry value is not a substring of the entry data.'
                    Entry.Value = Entry.Data
                else:
                    Entry.Value = Data

            elif TypeCode in self.PRIMITIVE:
                Type = self.PRIMITIVE[TypeCode]
                Entry.TypeName = repr(Type)
                package = StreamReader(Entry.Data).expect_with_meta(Type)
                Entry.Value = unpack(package)
            else:
                Entry.TypeName = 'UNKNOWN TYPE 0x{:X}'.format(TypeCode)


class NetStructuredResources(list):
    def __init__(self, data):
        list.__init__(self, NetManifestResource(StreamReader(data)).Resources)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Implements key derivation routines. These are mostly meant to be used as
modifiers for multibin expressions that can be passed as key arguments to
modules in `refinery.units.crypto.cipher`.
"""
from ... import Unit
from ....lib.argformats import multibin, number

try:
    from Crypto.Hash import SHA as SHA1
except ImportError:
    from Crypto.Hash import SHA1

from enum import Enum
from Crypto.Hash import MD2, MD4, MD5, SHA256, SHA512, SHA224, SHA384


__all__ = ['HashAlgorithms', 'KeyDerivation']


class HashAlgorithms(Enum):
    MD2 = MD2
    MD4 = MD4
    MD5 = MD5
    SHA256 = SHA256
    SHA512 = SHA512
    SHA224 = SHA224
    SHA384 = SHA384


class KeyDerivation(Unit, abstract=True):
    _DEFAULT_SALT = NotImplemented
    _DEFAULT_HASH = NotImplemented
    _DEFAULT_ITER = NotImplemented
    _DEFAULT_SIZE = None

    def interface(self, argp):
        kw = dict(
            type=number,
            help='The number of bytes to generate.'
        )
        if self._DEFAULT_SIZE is not None:
            hlp = F'The default is {self._DEFAULT_SIZE}.'
            kw.update(dict(
                default=self._DEFAULT_SIZE,
                nargs='?',
                help=F'{kw["help"]} {hlp}'
            ))
        argp.add_argument('size', **kw)

        if self._DEFAULT_SALT is not NotImplemented:
            kw = dict(
                type=multibin,
                help='Salt for derivation.'
            )
            if self._DEFAULT_SALT is not None:
                hlp = F'Default value is {self._DEFAULT_SALT.hex()} (hex).'
                kw.update(dict(
                    default=self._DEFAULT_SALT,
                    nargs='?',
                    help=F'{kw["help"]} {hlp}'
                ))
            argp.add_argument('salt', **kw)
        if self._DEFAULT_ITER is not NotImplemented:
            hlp = '' if not self._DEFAULT_ITER else F'Default is {self._DEFAULT_ITER}.'
            argp.add_argument(
                '-I', '--iterations',
                metavar='K',
                type=number,
                default=self._DEFAULT_ITER,
                help=F'Optionally specify the number of iterations. {hlp}'
            )
        if self._DEFAULT_HASH is not NotImplemented:
            argp.add_argument(
                '-A', '--algorithm',
                default=self._DEFAULT_HASH,
                metavar='A',
                choices=[h.name for h in HashAlgorithms],
                help=(
                    F'Optionally specify the hash algorithm A to use; A may be '
                    F'one of {", ".join(h.name for h in HashAlgorithms)}.'
                )
            )
        return super().interface(argp)

    @property
    def algorithm(self):
        name = getattr(self.args, 'algorithm', None)
        return name and HashAlgorithms[name].value

    @property
    def iterations(self):
        return getattr(self.args, 'iterations', None)

    @property
    def salt(self):
        return getattr(self.args, 'salt', None)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
try:
    import javaobj.v2 as java
except ImportError:
    java = None

import json

from .. import Unit

class JavaEncoder(json.JSONEncoder):

    def encode(self, obj):
        if isinstance(obj, dict):
            obj = {str(key): value for key, value in obj.items()}
        return super().encode(obj)

    def default(self, obj):
        try:
            return super().default(obj)
        except TypeError:
            if isinstance(obj, java.beans.JavaString):
                return str(obj)
            raise

class dsjava(Unit):
    """
    Deserialize Java serialized data and re-serialize as JSON.
    """
    def process(self, data):
        return json.dumps(
            java.loads(data),
            indent=4,
            cls=JavaEncoder
        ).encode(self.codec)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from .. import Unit
from ...lib.argformats import sliceobj


class snip(Unit):
    """
    Snips the input data based on a Python slice expression. For example,
    the initialization `slice 0::1 1::1` would yield a unit that first extracts
    every byte at an even position and then, every byte at an odd position. In
    this case, multiple outputs are produced.
    """
    def interface(self, argp):
        argp.add_argument(
            'slice',
            type=sliceobj,
            nargs='*',
            default=[slice(None, None)],
            help='Specify start:stop:step in Python slice syntax.'
        )
        return super().interface(argp)

    def process(self, data):
        for bounds in self.args.slice:
            yield data[bounds]
